from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class BAE_Normal_Map_Preprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.BAE_Normal_Map_Preprocessor"


class InpaintPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to inpaint.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(
        default=Mask(type="comfy.mask"), description="The mask to use for inpainting."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.InpaintPreprocessor"


from nodetool.common.comfy_node import EnableDisable


class LeReSDepthMapPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    rm_nearest: float | GraphNode | tuple[GraphNode, str] = Field(
        default=0.0, description="The nearest depth to remove."
    )
    rm_background: float | GraphNode | tuple[GraphNode, str] = Field(
        default=0.0, description="The background depth to remove."
    )
    boost: EnableDisable | GraphNode | tuple[GraphNode, str] = Field(
        default=EnableDisable("disable"), description="Whether to boost the depth map."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.LeReSDepthMapPreprocessor"


class MIDASDepthMapPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    a: float | GraphNode | tuple[GraphNode, str] = Field(
        default=6.283185307179586,
        description="Parameter 'a' for the MIDAS Depth Map Preprocessor.",
    )
    bg_threshold: float | GraphNode | tuple[GraphNode, str] = Field(
        default=0.1,
        description="Background threshold for the MIDAS Depth Map Preprocessor.",
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.MIDASDepthMapPreprocessor"


class MIDASNormalMapPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    a: float | GraphNode | tuple[GraphNode, str] = Field(
        default=6.283185307179586,
        description="Parameter 'a' for the MIDAS Normal Map Preprocessor.",
    )
    bg_threshold: float | GraphNode | tuple[GraphNode, str] = Field(
        default=0.1,
        description="Background threshold for the MIDAS Normal Map Preprocessor.",
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.MIDASNormalMapPreprocessor"


class MeshGraphormerDepthMapPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    mask_bbox_padding: int | GraphNode | tuple[GraphNode, str] = Field(
        default=30, description="The padding for the mask bounding box."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.MeshGraphormerDepthMapPreprocessor"


class PreprocessImage(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.PreprocessImage"


class ZoeDepthMapPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.normal_and_depth.ZoeDepthMapPreprocessor"



#https://github.com/siyeong0/Anime-Face-Segmentation/blob/main/network.py
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from torchvision.models import MobileNet_V2_Weights

class UNet(nn.Module):
    def __init__(self):
        super(UNet, self).__init__()
        self.NUM_SEG_CLASSES = 7 # Background, hair, face, eye, mouth, skin, clothes
        
        mobilenet_v2 = torchvision.models.mobilenet_v2(weights=MobileNet_V2_Weights.IMAGENET1K_V1)
        mob_blocks = mobilenet_v2.features
        
        # Encoder
        self.en_block0 = nn.Sequential(    # in_ch=3 out_ch=16
            mob_blocks[0],
            mob_blocks[1]
        )
        self.en_block1 = nn.Sequential(    # in_ch=16 out_ch=24
            mob_blocks[2],
            mob_blocks[3],
        )
        self.en_block2 = nn.Sequential(    # in_ch=24 out_ch=32
            mob_blocks[4],
            mob_blocks[5],
            mob_blocks[6],
        )
        self.en_block3 = nn.Sequential(    # in_ch=32 out_ch=96
            mob_blocks[7],
            mob_blocks[8],
            mob_blocks[9],
            mob_blocks[10],
            mob_blocks[11],
            mob_blocks[12],
            mob_blocks[13],
        )
        self.en_block4 = nn.Sequential(    # in_ch=96 out_ch=160
            mob_blocks[14],
            mob_blocks[15],
            mob_blocks[16],
        )
        
        # Decoder
        self.de_block4 = nn.Sequential(     # in_ch=160 out_ch=96
            nn.UpsamplingNearest2d(scale_factor=2),
            nn.Conv2d(160, 96, kernel_size=3, padding=1),
            nn.InstanceNorm2d(96),
            nn.LeakyReLU(0.1),
            nn.Dropout(p=0.2)
        )
        self.de_block3 = nn.Sequential(     # in_ch=96x2 out_ch=32
            nn.UpsamplingNearest2d(scale_factor=2),
            nn.Conv2d(96*2, 32, kernel_size=3, padding=1),
            nn.InstanceNorm2d(32),
            nn.LeakyReLU(0.1),
            nn.Dropout(p=0.2)
        )
        self.de_block2 = nn.Sequential(     # in_ch=32x2 out_ch=24
            nn.UpsamplingNearest2d(scale_factor=2),
            nn.Conv2d(32*2, 24, kernel_size=3, padding=1),
            nn.InstanceNorm2d(24),
            nn.LeakyReLU(0.1),
            nn.Dropout(p=0.2)
        )
        self.de_block1 = nn.Sequential(     # in_ch=24x2 out_ch=16
            nn.UpsamplingNearest2d(scale_factor=2),
            nn.Conv2d(24*2, 16, kernel_size=3, padding=1),
            nn.InstanceNorm2d(16),
            nn.LeakyReLU(0.1),
            nn.Dropout(p=0.2)
        )
        
        self.de_block0 = nn.Sequential(     # in_ch=16x2 out_ch=7
            nn.UpsamplingNearest2d(scale_factor=2),
            nn.Conv2d(16*2, self.NUM_SEG_CLASSES, kernel_size=3, padding=1),
            nn.Softmax2d()
        )
        
    def forward(self, x):
        e0 = self.en_block0(x)
        e1 = self.en_block1(e0)
        e2 = self.en_block2(e1)
        e3 = self.en_block3(e2)
        e4 = self.en_block4(e3)
        
        d4 = self.de_block4(e4)
        c4 = torch.cat((d4,e3),1)
        d3 = self.de_block3(c4)
        c3 = torch.cat((d3,e2),1)
        d2 = self.de_block2(c3)
        c2 =torch.cat((d2,e1),1)
        d1 = self.de_block1(c2)
        c1 = torch.cat((d1,e0),1)
        y = self.de_block0(c1)
        
        return y


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.replicate.audio.transcribe import Task
from nodetool.nodes.replicate.audio.transcribe import Language
from nodetool.nodes.replicate.audio.transcribe import Timestamp

class IncrediblyFastWhisper(GraphNode):
    task: Task | GraphNode | tuple[GraphNode, str] = Field(default='transcribe', description='Task to perform: transcribe or translate to another language.')
    audio: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None), description='Audio file')
    hf_token: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description="Provide a hf.co/settings/token for Pyannote.audio to diarise the audio clips. You need to agree to the terms in 'https://huggingface.co/pyannote/speaker-diarization-3.1' and 'https://huggingface.co/pyannote/segmentation-3.0' first.")
    language: Language | GraphNode | tuple[GraphNode, str] = Field(default='None', description="Language spoken in the audio, specify 'None' to perform language detection.")
    timestamp: Timestamp | GraphNode | tuple[GraphNode, str] = Field(default='chunk', description='Whisper supports both chunked as well as word level timestamps.')
    batch_size: int | GraphNode | tuple[GraphNode, str] = Field(default=24, description='Number of parallel batches you want to compute. Reduce if you face OOMs.')
    diarise_audio: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Use Pyannote.audio to diarise the audio clips. You will need to provide hf_token below too.')
    @classmethod
    def get_node_type(cls): return "replicate.audio.transcribe.IncrediblyFastWhisper"


from nodetool.nodes.replicate.audio.transcribe import Language
from nodetool.nodes.replicate.audio.transcribe import Transcription

class Whisper(GraphNode):
    audio: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None), description='Audio file')
    model: str | GraphNode | tuple[GraphNode, str] = Field(default='large-v3', description='This version only supports Whisper-large-v3.')
    language: Language | GraphNode | tuple[GraphNode, str] = Field(default=None, description='language spoken in the audio, specify None to perform language detection')
    patience: float | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='optional patience value to use in beam decoding, as in https://arxiv.org/abs/2204.05424, the default (1.0) is equivalent to conventional beam search')
    translate: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Translate the text to English when set to True')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='temperature to use for sampling')
    transcription: Transcription | GraphNode | tuple[GraphNode, str] = Field(default='plain text', description='Choose the format for the transcription')
    initial_prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='optional text to provide as a prompt for the first window.')
    suppress_tokens: str | GraphNode | tuple[GraphNode, str] = Field(default='-1', description="comma-separated list of token ids to suppress during sampling; '-1' will suppress most special characters except common punctuations")
    logprob_threshold: float | GraphNode | tuple[GraphNode, str] = Field(default=-1, description='if the average log probability is lower than this value, treat the decoding as failed')
    no_speech_threshold: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='if the probability of the <|nospeech|> token is higher than this value AND the decoding has failed due to `logprob_threshold`, consider the segment as silence')
    condition_on_previous_text: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='if True, provide the previous output of the model as a prompt for the next window; disabling may make the text inconsistent across windows, but the model becomes less prone to getting stuck in a failure loop')
    compression_ratio_threshold: float | GraphNode | tuple[GraphNode, str] = Field(default=2.4, description='if the gzip compression ratio is higher than this value, treat the decoding as failed')
    temperature_increment_on_fallback: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='temperature to increase when falling back when the decoding fails to meet either of the thresholds below')
    @classmethod
    def get_node_type(cls): return "replicate.audio.transcribe.Whisper"





from nodetool.api.types.graph import Edge, Node
import uuid

from nodetool.workflows.base_node import get_node_class, get_node_class_by_name


def read_graph(json: dict) -> tuple[list[Edge], list[Node]]:
    """
    This function reads a graph from a dictionary representation.

    The format of the dictionary representation has the following structure:
    {
        "source_node_id": {
            "type": "node_type",
        },
        "node_id": {
            "type": "node_type",
            "data": {
                "property_name": "property_value",
                "property_name": ["source_node_id", "source_node_output_handle"]
            }
        }
    }

    It also supports the comfy workflow format:
    {
        "source_node_id": {
            "class_type": "node_type",
        },
        "node_id": {
            "class_type": "node_type",
            "inputs": {
                "property_name": "property_value",
                "property_name": ["source_node_id", 0]
            }
        }
    }

    Parameters:
    - json (dict): The dictionary representation of the graph.

    Returns:
    - A tuple containing two lists: a list of edges and a list of nodes.

    Raises:
    - Exception: If the node type or class cannot be found.
    - Exception: If a source node referenced by a node cannot be found.

    Example usage:
    ```
    graph_json = {
        "source_node_id": {
            "type": "node_type",
        },
        "node_id": {
            "type": "node_type",
            "data": {
                "property_name": "property_value",
                "property_name": ["source_node_id", "source_node_output_handle"]
            }
        }
    }
    edges, nodes = read_graph(graph_json)
    ```
    """
    nodes: list[Node] = []
    edges: list[Edge] = []
    node_by_id: dict[str, Node] = {}

    def generate_edge_id():
        """
        Finds the highest ID in the list of edges and returns a new ID that is one higher.
        """
        return str(max([int(edge.id or "0", 16) for edge in edges], default=0) + 1)

    # we loop through the nodes twice to make sure all nodes are created before we try to connect them
    for node_id, node in json.items():
        # supporting comfy workflow format
        if "class_type" in node:
            classes = get_node_class_by_name(node["class_type"])
            assert (
                len(classes) > 0
            ), f"Could not find node class {node['class_type']} for node {node_id}"
            node_type = classes[0].get_node_type()
            if len(classes) > 1:
                for cls in classes:
                    if cls.get_node_type().startswith("comfy."):
                        node_type = cls.get_node_type()
                        break

        elif "type" in node:
            assert get_node_class(
                node["type"]
            ), f"Could not find node class {node['type']} for node {node_id}"
            node_type = node["type"]
        else:
            raise Exception(f"Node {node_id} does not have a type")

        node_inputs = node.get("data", node.get("inputs", {}))
        data = {
            name: value
            for name, value in node_inputs.items()
            if not isinstance(value, list)
        }

        parent_id = node.get("parent_id", None)
        ui_properties = {}

        if "position" in node:
            ui_properties["position"] = node["position"]

        if "width" in node:
            ui_properties["width"] = node["width"]

        if "height" in node:
            ui_properties["height"] = node["height"]

        node_instance = Node(
            id=node_id,
            type=node_type,
            data=data,
            parent_id=parent_id,
            ui_properties=ui_properties,
        )

        nodes.append(node_instance)
        node_by_id[node_id] = node_instance

    # connect the nodes
    for node_id, node in json.items():
        # supporting comfy workflow format
        node_inputs = node.get("data", node.get("inputs", {}))

        # loop through the inputs and connect them
        for input_name, input_value in node_inputs.items():
            # a list value means that the input is connected to an output
            # the list contains the source node ID and the output handle
            if isinstance(input_value, list):
                source_id, source_handle = input_value

                # look up the source node by ID, so we can find the output handle
                source_node = node_by_id.get(source_id)
                if source_node is None:
                    raise Exception(
                        f"Could not find source node {source_id} referenced by {node_id}"
                    )

                # supporting comfy workflow format, which uses index instead of handle
                # the index will be resolved in the workflow runner
                if isinstance(source_handle, int):
                    source_node = node_by_id.get(source_id)
                    assert source_node
                    source_node_class = get_node_class(source_node.type)
                    assert source_node_class
                    source_handle = str(
                        source_node_class.find_output_by_index(source_handle).name
                    )

                edge = Edge(
                    id=generate_edge_id(),
                    source=source_id,
                    sourceHandle=source_handle,
                    target=node_id,
                    targetHandle=input_name,
                )

                edges.append(edge)

    return edges, nodes



'''
Copyright 2017 Javier Romero, Dimitrios Tzionas, Michael J Black and the Max Planck Gesellschaft.  All rights reserved.
This software is provided for research purposes only.
By using this software you agree to the terms of the MANO/SMPL+H Model license here http://mano.is.tue.mpg.de/license

More information about MANO/SMPL+H is available at http://mano.is.tue.mpg.de.
For comments or questions, please email us at: mano@tue.mpg.de


About this file:
================
This file defines a wrapper for the loading functions of the MANO model.

Modules included:
- load_model:
  loads the MANO model from a given file location (i.e. a .pkl file location),
  or a dictionary object.

'''


import numpy as np
import mano.webuser.lbs as lbs
from mano.webuser.posemapper import posemap
import scipy.sparse as sp


def ischumpy(x):
    return hasattr(x, 'dterms')


def verts_decorated(trans,
                    pose,
                    v_template,
                    J_regressor,
                    weights,
                    kintree_table,
                    bs_style,
                    f,
                    bs_type=None,
                    posedirs=None,
                    betas=None,
                    shapedirs=None,
                    want_Jtr=False):

    for which in [
            trans, pose, v_template, weights, posedirs, betas, shapedirs
    ]:
        if which is not None:
            assert ischumpy(which)

    v = v_template

    if shapedirs is not None:
        if betas is None:
            betas = np.zeros(shapedirs.shape[-1])
        v_shaped = v + shapedirs.dot(betas)
    else:
        v_shaped = v

    if posedirs is not None:
        v_posed = v_shaped + posedirs.dot(posemap(bs_type)(pose))
    else:
        v_posed = v_shaped

    v = v_posed

    if sp.issparse(J_regressor):
        J_tmpx = np.matmul(J_regressor, v_shaped[:, 0])
        J_tmpy = np.matmul(J_regressor, v_shaped[:, 1])
        J_tmpz = np.matmul(J_regressor, v_shaped[:, 2])
        J = np.vstack((J_tmpx, J_tmpy, J_tmpz)).T
    else:
        assert (ischumpy(J))

    assert (bs_style == 'lbs')
    result, Jtr = lbs.verts_core(
        pose, v, J, weights, kintree_table, want_Jtr=True, xp=np)

    tr = trans.reshape((1, 3))
    result = result + tr
    Jtr = Jtr + tr

    result.trans = trans
    result.f = f
    result.pose = pose
    result.v_template = v_template
    result.J = J
    result.J_regressor = J_regressor
    result.weights = weights
    result.kintree_table = kintree_table
    result.bs_style = bs_style
    result.bs_type = bs_type
    if posedirs is not None:
        result.posedirs = posedirs
        result.v_posed = v_posed
    if shapedirs is not None:
        result.shapedirs = shapedirs
        result.betas = betas
        result.v_shaped = v_shaped
    if want_Jtr:
        result.J_transformed = Jtr
    return result


def verts_core(pose,
               v,
               J,
               weights,
               kintree_table,
               bs_style,
               want_Jtr=False,
               xp=np):
    
    assert (bs_style == 'lbs')
    result = lbs.verts_core(pose, v, J, weights, kintree_table, want_Jtr, xp)
    return result


import cv2
import numpy as np
import torch
from scipy.ndimage.filters import gaussian_filter
from skimage.measure import label

from . import util
from .model import handpose_model


class Hand(object):
    def __init__(self, model_path):
        self.model = handpose_model()
        model_dict = util.transfer(self.model, torch.load(model_path))
        self.model.load_state_dict(model_dict)
        self.model.eval()

    def to(self, device):
        self.model.to(device)
        return self

    def __call__(self, oriImgRaw):
        device = next(iter(self.model.parameters())).device
        scale_search = [0.5, 1.0, 1.5, 2.0]
        # scale_search = [0.5]
        boxsize = 368
        stride = 8
        padValue = 128
        thre = 0.05
        multiplier = [x * boxsize for x in scale_search]

        wsize = 128
        heatmap_avg = np.zeros((wsize, wsize, 22))

        Hr, Wr, Cr = oriImgRaw.shape

        oriImg = cv2.GaussianBlur(oriImgRaw, (0, 0), 0.8)

        for m in range(len(multiplier)):
            scale = multiplier[m]
            imageToTest = util.smart_resize(oriImg, (scale, scale))

            imageToTest_padded, pad = util.padRightDownCorner(imageToTest, stride, padValue)
            im = np.transpose(np.float32(imageToTest_padded[:, :, :, np.newaxis]), (3, 2, 0, 1)) / 256 - 0.5
            im = np.ascontiguousarray(im)

            data = torch.from_numpy(im).float()
            data = data.to(device)

            with torch.no_grad():
                output = self.model(data).cpu().numpy()

            # extract outputs, resize, and remove padding
            heatmap = np.transpose(np.squeeze(output), (1, 2, 0))  # output 1 is heatmaps
            heatmap = util.smart_resize_k(heatmap, fx=stride, fy=stride)
            heatmap = heatmap[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
            heatmap = util.smart_resize(heatmap, (wsize, wsize))

            heatmap_avg += heatmap / len(multiplier)

        all_peaks = []
        for part in range(21):
            map_ori = heatmap_avg[:, :, part]
            one_heatmap = gaussian_filter(map_ori, sigma=3)
            binary = np.ascontiguousarray(one_heatmap > thre, dtype=np.uint8)

            if np.sum(binary) == 0:
                all_peaks.append([0, 0])
                continue
            label_img, label_numbers = label(binary, return_num=True, connectivity=binary.ndim)
            max_index = np.argmax([np.sum(map_ori[label_img == i]) for i in range(1, label_numbers + 1)]) + 1
            label_img[label_img != max_index] = 0
            map_ori[label_img == 0] = 0

            y, x = util.npmax(map_ori)
            y = int(float(y) * float(Hr) / float(wsize))
            x = int(float(x) * float(Wr) / float(wsize))
            all_peaks.append([x, y])
        return np.array(all_peaks)

if __name__ == "__main__":
    hand_estimation = Hand('../model/hand_pose_model.pth')

    # test_image = '../images/hand.jpg'
    test_image = '../images/hand.jpg'
    oriImg = cv2.imread(test_image)  # B,G,R order
    peaks = hand_estimation(oriImg)
    canvas = util.draw_handpose(oriImg, peaks, True)
    cv2.imshow('', canvas)
    cv2.waitKey(0)


from enum import Enum
from pydantic import Field
from nodetool.metadata.types import ImageRef
from nodetool.common.huggingface_node import HuggingfaceNode
from nodetool.workflows.processing_context import ProcessingContext


class ModelId(str, Enum):
    GOOGLE_VIT_BASE_PATCH16_224 = "google/vit-base-patch16-224"
    MICROSOFT_RESNET_50 = "microsoft/resnet-50"
    MICROSOFT_RESNET_18 = "microsoft/resnet-18"
    APPLE_MOBILEVIT_SMALL = "apple/mobilevit-small"
    NATERAW_VIT_AGE_CLASSIFIER = "nateraw/vit-age-classifier"
    FALCONSAI_NSFW_IMAGE_DETECTION = "Falconsai/nsfw_image_detection"
    MICROSOFT_BEIT_BASE_PATCH16_224_PT22K_FT22K = (
        "microsoft/beit-base-patch16-224-pt22k-ft22k"
    )
    TIMM_VIT_LARGE_PATCH14_CLIP_224_OPENAI_FT_IN12K_IN1K = (
        "timm/vit_large_patch14_clip_224.openai_ft_in12k_in1k"
    )
    ORGANIKA_SDXL_DETECTOR = "Organika/sdxl-detector"
    RIZVANDWIKI_GENDER_CLASSIFICATION_2 = "rizvandwiki/gender-classification-2"


class Classifier(HuggingfaceNode):
    """
    Image classification is the task of assigning a label or class to an entire image.
    Images are expected to have only one class for each image. Image classification models
    take an image as input and return a prediction about which class the image belongs to.
    """

    model: ModelId = Field(
        default=ModelId.GOOGLE_VIT_BASE_PATCH16_224,
        title="Model ID on Huggingface",
        description="The model ID to use for the classification",
    )
    image: ImageRef = Field(
        default=ImageRef(),
        title="Image",
        description="The input image to classify",
    )

    async def process(self, context: ProcessingContext) -> dict[str, float]:
        image = await context.asset_to_io(self.image)
        result = await self.run_huggingface(
            model_id=self.model.value, context=context, data=image.read()
        )
        return result[0]



from base64 import b64decode
import PIL.Image
from io import BytesIO
from nodetool.common.environment import Environment
from nodetool.common.openai_nodes import calculate_cost
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import ImageRef
from nodetool.workflows.base_node import BaseNode
from pydantic import Field
from enum import Enum

from openai.types.images_response import ImagesResponse


class Dall_E(BaseNode):
    """
    Generates images from textual descriptions using DALLE-3.
    image, t2i, tti, text-to-image, create, generate, dall-e, picture, photo, art, drawing, illustration
    Returns an image based on the provided textual prompt.
    """

    class Size(str, Enum):
        _1024x1024 = "1024x1024"
        _1792x1024 = "1792x1024"
        _1024x1792 = "1024x1792"

    class Style(str, Enum):
        vivid = "vivid"
        natural = "natural"

    class Quality(str, Enum):
        standard = "standard"
        hd = "hd"

    prompt: str = Field(default="", description="The prompt to use.")
    size: Size = Field(
        default=Size._1024x1024, description="The size of the image to generate."
    )
    quality: Quality = Field(
        default=Quality.standard, description="The quality of the image to generate."
    )
    style: Style = Field(default=Style.natural, description="The style to use.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        response = await context.run_prediction(
            node_id=self._id,
            provider="openai",
            model="dall-e-3",
            params={
                "prompt": self.prompt,
                "style": self.style.value,
                "size": self.size.value,
                "quality": self.quality.value,
            },
        )
        image = ImagesResponse(**response.json())

        assert len(image.data) > 0
        b64 = image.data[0].b64_json
        assert b64 is not None
        file = b64decode(b64)
        pil_image = PIL.Image.open(BytesIO(file))
        return await context.image_from_pil(pil_image)



import json
from pydantic import BaseModel, Field
import uuid

from nodetool.api.types.graph import Graph, Node
from nodetool.common.environment import Environment
from nodetool.workflows.run_job_request import RunJobRequest
from nodetool.workflows.run_workflow import run_workflow


class GraphNode(BaseModel):
    """
    Represents a node in a graph DSL.
    """

    id: str = Field(default_factory=lambda: str(uuid.uuid4()))

    @classmethod
    def set_node_type(cls, node_type: str):
        cls.__annotations__["node_type"] = node_type

    @classmethod
    def get_node_type(cls) -> str:
        return cls.__annotations__["node_type"]


def graph(*nodes):
    """
    Create a graph representation based on the given nodes.

    Args:
      *nodes: Variable number of nodes to be included in the graph.

    Returns:
      Graph: A graph object containing the nodes and edges.

    """
    from nodetool.dsl.graph_node_converter import GraphNodeConverter

    g = GraphNodeConverter()
    for node in nodes:
        g.add(node)
    nodes = [
        Node(
            id=n._id,
            type=n.get_node_type(),
            data=n.model_dump(),
        )
        for n in g.nodes.values()
    ]
    return Graph(nodes=nodes, edges=g.edges)


def run(graph: Graph, user_id: str = "1", auth_token: str = "token"):
    """
    Run the workflow with the given graph.

    Args:
      graph (Graph): The graph object representing the workflow.

    Returns:
      Any: The result of the workflow execution.
    """
    req = RunJobRequest(user_id=user_id, auth_token=auth_token, graph=graph)

    res = None
    for str_msg in run_workflow(req):
        msg = json.loads(str_msg)
        if msg["type"] == "workflow_update":
            res = msg["result"]
    return res



from datetime import datetime
from nodetool.models.base_model import DBModel, DBField, create_time_ordered_uuid


class Thread(DBModel):
    @classmethod
    def get_table_schema(cls):
        return {
            "table_name": "nodetool_threads",
            "key_schema": {"id": "HASH"},
            "attribute_definitions": {"id": "S", "user_id": "S"},
            "global_secondary_indexes": {
                "nodetool_thread_user_index": {"user_id": "HASH"},
            },
        }

    id: str = DBField(hash_key=True)
    user_id: str = DBField(default="")
    created_at: datetime = DBField(default_factory=datetime.now)
    updated_at: datetime = DBField(default_factory=datetime.now)

    @classmethod
    def find(cls, user_id: str, id: str):
        thread = cls.get(id)
        if thread and thread.user_id == user_id:
            return thread
        return None

    @classmethod
    def create(cls, user_id: str, **kwargs) -> "Thread":
        return super().create(
            id=create_time_ordered_uuid(),
            user_id=user_id,
            **kwargs,
        )

    @classmethod
    def paginate(
        cls,
        user_id: str,
        limit: int = 10,
        start_key: str | None = None,
        reverse: bool = False,
    ):
        return cls.query(
            condition="user_id = :user_id",
            values={":user_id": user_id},
            index="nodetool_thread_user_index",
            limit=limit,
            start_key=start_key,
            reverse=reverse,
        )



"""
Rendering tools for 3D mesh visualization on 2D image.

Parts of the code are taken from https://github.com/akanazawa/hmr
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import cv2
import code
from opendr.camera import ProjectPoints
from opendr.renderer import ColoredRenderer, TexturedRenderer
from opendr.lighting import LambertianPointLight
import random


# Rotate the points by a specified angle.
def rotateY(points, angle):
    ry = np.array([
        [np.cos(angle), 0., np.sin(angle)], [0., 1., 0.],
        [-np.sin(angle), 0., np.cos(angle)]
    ])
    return np.dot(points, ry)

def draw_skeleton(input_image, joints, draw_edges=True, vis=None, radius=None):
    """
    joints is 3 x 19. but if not will transpose it.
    0: Right ankle
    1: Right knee
    2: Right hip
    3: Left hip
    4: Left knee
    5: Left ankle
    6: Right wrist
    7: Right elbow
    8: Right shoulder
    9: Left shoulder
    10: Left elbow
    11: Left wrist
    12: Neck
    13: Head top
    14: nose
    15: left_eye
    16: right_eye
    17: left_ear
    18: right_ear
    """

    if radius is None:
        radius = max(4, (np.mean(input_image.shape[:2]) * 0.01).astype(int))

    colors = {
        'pink': (197, 27, 125),  # L lower leg
        'light_pink': (233, 163, 201),  # L upper leg
        'light_green': (161, 215, 106),  # L lower arm
        'green': (77, 146, 33),  # L upper arm
        'red': (215, 48, 39),  # head
        'light_red': (252, 146, 114),  # head
        'light_orange': (252, 141, 89),  # chest
        'purple': (118, 42, 131),  # R lower leg
        'light_purple': (175, 141, 195),  # R upper
        'light_blue': (145, 191, 219),  # R lower arm
        'blue': (69, 117, 180),  # R upper arm
        'gray': (130, 130, 130),  #
        'white': (255, 255, 255),  #
    }

    image = input_image.copy()
    input_is_float = False

    if np.issubdtype(image.dtype, np.float):
        input_is_float = True
        max_val = image.max()
        if max_val <= 2.:  # should be 1 but sometimes it's slightly above 1
            image = (image * 255).astype(np.uint8)
        else:
            image = (image).astype(np.uint8)

    if joints.shape[0] != 2:
        joints = joints.T
    joints = np.round(joints).astype(int)

    jcolors = [
        'light_pink', 'light_pink', 'light_pink', 'pink', 'pink', 'pink',
        'light_blue', 'light_blue', 'light_blue', 'blue', 'blue', 'blue',
        'purple', 'purple', 'red', 'green', 'green', 'white', 'white',
        'purple', 'purple', 'red', 'green', 'green', 'white', 'white'
    ]

    if joints.shape[1] == 19:
        # parent indices -1 means no parents
        parents = np.array([
            1, 2, 8, 9, 3, 4, 7, 8, 12, 12, 9, 10, 14, -1, 13, -1, -1, 15, 16
        ])
        # Left is light and right is dark
        ecolors = {
            0: 'light_pink',
            1: 'light_pink',
            2: 'light_pink',
            3: 'pink',
            4: 'pink',
            5: 'pink',
            6: 'light_blue',
            7: 'light_blue',
            8: 'light_blue',
            9: 'blue',
            10: 'blue',
            11: 'blue',
            12: 'purple',
            17: 'light_green',
            18: 'light_green',
            14: 'purple'
        }
    elif joints.shape[1] == 14:
        parents = np.array([
            1,
            2,
            8,
            9,
            3,
            4,
            7,
            8,
            -1,
            -1,
            9,
            10,
            13,
            -1,
        ])
        ecolors = {
            0: 'light_pink',
            1: 'light_pink',
            2: 'light_pink',
            3: 'pink',
            4: 'pink',
            5: 'pink',
            6: 'light_blue',
            7: 'light_blue',
            10: 'light_blue',
            11: 'blue',
            12: 'purple'
        }
    elif joints.shape[1] == 21:  # hand
        parents = np.array([
            -1,
            0,
            1,
            2,
            3,
            0,
            5,
            6,
            7,
            0,
            9,
            10,
            11,
            0,
            13,
            14,
            15,
            0,
            17,
            18,
            19,
        ])
        ecolors = {
            0: 'light_purple',
            1: 'light_green',
            2: 'light_green',
            3: 'light_green',
            4: 'light_green',
            5: 'pink',
            6: 'pink',
            7: 'pink',
            8: 'pink',
            9: 'light_blue',
            10: 'light_blue',
            11: 'light_blue',
            12: 'light_blue',
            13: 'light_red',
            14: 'light_red',
            15: 'light_red',
            16: 'light_red',
            17: 'purple',
            18: 'purple',
            19: 'purple',
            20: 'purple',
        }
    else:
        print('Unknown skeleton!!')

    for child in range(len(parents)):
        point = joints[:, child]
        # If invisible skip
        if vis is not None and vis[child] == 0:
            continue
        if draw_edges:
            cv2.circle(image, (point[0], point[1]), radius, colors['white'],
                       -1)
            cv2.circle(image, (point[0], point[1]), radius - 1,
                       colors[jcolors[child]], -1)
        else:
            # cv2.circle(image, (point[0], point[1]), 5, colors['white'], 1)
            cv2.circle(image, (point[0], point[1]), radius - 1,
                       colors[jcolors[child]], 1)
            # cv2.circle(image, (point[0], point[1]), 5, colors['gray'], -1)
        pa_id = parents[child]
        if draw_edges and pa_id >= 0:
            if vis is not None and vis[pa_id] == 0:
                continue
            point_pa = joints[:, pa_id]
            cv2.circle(image, (point_pa[0], point_pa[1]), radius - 1,
                       colors[jcolors[pa_id]], -1)
            if child not in ecolors.keys():
                print('bad')
                import ipdb
                ipdb.set_trace()
            cv2.line(image, (point[0], point[1]), (point_pa[0], point_pa[1]),
                     colors[ecolors[child]], radius - 2)

    # Convert back in original dtype
    if input_is_float:
        if max_val <= 1.:
            image = image.astype(np.float32) / 255.
        else:
            image = image.astype(np.float32)

    return image

def draw_text(input_image, content):
    """
    content is a dict. draws key: val on image
    Assumes key is str, val is float
    """
    image = input_image.copy()
    input_is_float = False
    if np.issubdtype(image.dtype, np.float):
        input_is_float = True
        image = (image * 255).astype(np.uint8)

    black = (255, 255, 0)
    margin = 15
    start_x = 5
    start_y = margin
    for key in sorted(content.keys()):
        text = "%s: %.2g" % (key, content[key])
        cv2.putText(image, text, (start_x, start_y), 0, 0.45, black)
        start_y += margin

    if input_is_float:
        image = image.astype(np.float32) / 255.
    return image

def visualize_reconstruction(img, img_size, gt_kp, vertices, pred_kp, camera, renderer, color='pink', focal_length=1000):
    """Overlays gt_kp and pred_kp on img.
    Draws vert with text.
    Renderer is an instance of SMPLRenderer.
    """
    gt_vis = gt_kp[:, 2].astype(bool)
    loss = np.sum((gt_kp[gt_vis, :2] - pred_kp[gt_vis])**2)
    debug_text = {"sc": camera[0], "tx": camera[1], "ty": camera[2], "kpl": loss}
    # Fix a flength so i can render this with persp correct scale
    res = img.shape[1]
    camera_t = np.array([camera[1], camera[2], 2*focal_length/(res * camera[0] +1e-9)])
    rend_img = renderer.render(vertices, camera_t=camera_t,
                               img=img, use_bg=True,
                               focal_length=focal_length,
                               body_color=color)
    rend_img = draw_text(rend_img, debug_text)

    # Draw skeleton
    gt_joint = ((gt_kp[:, :2] + 1) * 0.5) * img_size
    pred_joint = ((pred_kp + 1) * 0.5) * img_size
    img_with_gt = draw_skeleton( img, gt_joint, draw_edges=False, vis=gt_vis)
    skel_img = draw_skeleton(img_with_gt, pred_joint)

    combined = np.hstack([skel_img, rend_img])

    return combined

def visualize_reconstruction_test(img, img_size, gt_kp, vertices, pred_kp, camera, renderer, score, color='pink', focal_length=1000):
    """Overlays gt_kp and pred_kp on img.
    Draws vert with text.
    Renderer is an instance of SMPLRenderer.
    """
    gt_vis = gt_kp[:, 2].astype(bool)
    loss = np.sum((gt_kp[gt_vis, :2] - pred_kp[gt_vis])**2)
    debug_text = {"sc": camera[0], "tx": camera[1], "ty": camera[2], "kpl": loss, "pa-mpjpe": score*1000}
    # Fix a flength so i can render this with persp correct scale
    res = img.shape[1]
    camera_t = np.array([camera[1], camera[2], 2*focal_length/(res * camera[0] +1e-9)])
    rend_img = renderer.render(vertices, camera_t=camera_t,
                               img=img, use_bg=True,
                               focal_length=focal_length,
                               body_color=color)
    rend_img = draw_text(rend_img, debug_text)

    # Draw skeleton
    gt_joint = ((gt_kp[:, :2] + 1) * 0.5) * img_size
    pred_joint = ((pred_kp + 1) * 0.5) * img_size
    img_with_gt = draw_skeleton( img, gt_joint, draw_edges=False, vis=gt_vis)
    skel_img = draw_skeleton(img_with_gt, pred_joint)

    combined = np.hstack([skel_img, rend_img])

    return combined



def visualize_reconstruction_and_att(img, img_size, vertices_full, vertices, vertices_2d, camera, renderer, ref_points, attention, focal_length=1000):
    """Overlays gt_kp and pred_kp on img.
    Draws vert with text.
    Renderer is an instance of SMPLRenderer.
    """
    # Fix a flength so i can render this with persp correct scale
    res = img.shape[1]
    camera_t = np.array([camera[1], camera[2], 2*focal_length/(res * camera[0] +1e-9)])
    rend_img = renderer.render(vertices_full, camera_t=camera_t,
                               img=img, use_bg=True, 
                               focal_length=focal_length, body_color='light_blue')


    heads_num, vertex_num, _ = attention.shape

    all_head = np.zeros((vertex_num,vertex_num))

    ###### find max
    # for i in range(vertex_num):
    #     for j in range(vertex_num):
    #         all_head[i,j] = np.max(attention[:,i,j])

    ##### find avg
    for h in range(4):
        att_per_img = attention[h]
        all_head = all_head + att_per_img   
    all_head = all_head/4

    col_sums = all_head.sum(axis=0)
    all_head = all_head / col_sums[np.newaxis, :]


    # code.interact(local=locals())

    combined = []
    if vertex_num>400:  # body
        selected_joints = [6,7,4,5,13] # [6,7,4,5,13,12] 
    else: # hand  
        selected_joints = [0, 4, 8, 12, 16, 20]
    # Draw attention
    for ii in range(len(selected_joints)):
        reference_id = selected_joints[ii]
        ref_point = ref_points[reference_id]
        attention_to_show = all_head[reference_id][14::] 
        min_v = np.min(attention_to_show)
        max_v = np.max(attention_to_show)
        norm_attention_to_show = (attention_to_show - min_v)/(max_v-min_v)

        vertices_norm = ((vertices_2d + 1) * 0.5) * img_size
        ref_norm = ((ref_point + 1) * 0.5) * img_size
        image = np.zeros_like(rend_img)

        for jj in range(vertices_norm.shape[0]):
            x = int(vertices_norm[jj,0])
            y = int(vertices_norm[jj,1])
            cv2.circle(image,(x,y), 1, (255,255,255), -1) 

        total_to_draw = []
        for jj in range(vertices_norm.shape[0]):
            thres = 0.0
            if norm_attention_to_show[jj]>thres:
                things = [norm_attention_to_show[jj], ref_norm, vertices_norm[jj]]
                total_to_draw.append(things)
                # plot_one_line(ref_norm, vertices_norm[jj], image, reference_id, alpha=0.4*(norm_attention_to_show[jj]-thres)/(1-thres)  )
        total_to_draw.sort()
        max_att_score = total_to_draw[-1][0]
        for item in total_to_draw:
            attention_score = item[0]
            ref_point = item[1]
            vertex = item[2]
            plot_one_line(ref_point, vertex, image, ii, alpha=(attention_score-thres)/(max_att_score-thres)  )
        # code.interact(local=locals())
        if len(combined)==0:
            combined = image
        else:
            combined = np.hstack([combined, image])

    final = np.hstack([img, combined, rend_img])

    return final


def visualize_reconstruction_and_att_local(img, img_size, vertices_full, vertices, vertices_2d, camera, renderer, ref_points, attention, color='light_blue', focal_length=1000):
    """Overlays gt_kp and pred_kp on img.
    Draws vert with text.
    Renderer is an instance of SMPLRenderer.
    """
    # Fix a flength so i can render this with persp correct scale
    res = img.shape[1]
    camera_t = np.array([camera[1], camera[2], 2*focal_length/(res * camera[0] +1e-9)])
    rend_img = renderer.render(vertices_full, camera_t=camera_t,
                               img=img, use_bg=True, 
                               focal_length=focal_length, body_color=color)
    heads_num, vertex_num, _ = attention.shape
    all_head = np.zeros((vertex_num,vertex_num))

    ##### compute avg attention for 4 attention heads
    for h in range(4):
        att_per_img = attention[h]
        all_head = all_head + att_per_img   
    all_head = all_head/4

    col_sums = all_head.sum(axis=0)
    all_head = all_head / col_sums[np.newaxis, :]

    combined = []
    if vertex_num>400:  # body
        selected_joints = [7]  # [6,7,4,5,13,12] 
    else: # hand  
        selected_joints = [0] # [0, 4, 8, 12, 16, 20] 
    # Draw attention
    for ii in range(len(selected_joints)):
        reference_id = selected_joints[ii]
        ref_point = ref_points[reference_id]
        attention_to_show = all_head[reference_id][14::] 
        min_v = np.min(attention_to_show)
        max_v = np.max(attention_to_show)
        norm_attention_to_show = (attention_to_show - min_v)/(max_v-min_v)
        vertices_norm = ((vertices_2d + 1) * 0.5) * img_size
        ref_norm = ((ref_point + 1) * 0.5) * img_size
        image = rend_img*0.4

        total_to_draw = []
        for jj in range(vertices_norm.shape[0]):
            thres = 0.0
            if norm_attention_to_show[jj]>thres:
                things = [norm_attention_to_show[jj], ref_norm, vertices_norm[jj]]
                total_to_draw.append(things)
        total_to_draw.sort()
        max_att_score = total_to_draw[-1][0]
        for item in total_to_draw:
            attention_score = item[0]
            ref_point = item[1]
            vertex = item[2]
            plot_one_line(ref_point, vertex, image, ii, alpha=(attention_score-thres)/(max_att_score-thres)  )

        for jj in range(vertices_norm.shape[0]):
            x = int(vertices_norm[jj,0])
            y = int(vertices_norm[jj,1])
            cv2.circle(image,(x,y), 1, (255,255,255), -1) 

        if len(combined)==0:
            combined = image
        else:
            combined = np.hstack([combined, image])

    final = np.hstack([img, combined, rend_img])

    return final


def visualize_reconstruction_no_text(img, img_size, vertices, camera, renderer, color='pink', focal_length=1000):
    """Overlays gt_kp and pred_kp on img.
    Draws vert with text.
    Renderer is an instance of SMPLRenderer.
    """
    # Fix a flength so i can render this with persp correct scale
    res = img.shape[1]
    camera_t = np.array([camera[1], camera[2], 2*focal_length/(res * camera[0] +1e-9)])
    rend_img = renderer.render(vertices, camera_t=camera_t,
                               img=img, use_bg=True,
                               focal_length=focal_length,
                               body_color=color)


    combined = np.hstack([img, rend_img])

    return combined


def plot_one_line(ref, vertex, img, color_index, alpha=0.0, line_thickness=None):
    # 13,6,7,8,3,4,5
    # att_colors = [(255, 221, 104), (255, 255, 0), (255, 215, 227),  (210, 240, 119), \
    #          (209, 238, 245), (244, 200, 243),  (233, 242, 216)] 
    att_colors = [(255, 255, 0), (244, 200, 243),  (210, 243, 119), (209, 238, 255), (200, 208, 255), (250, 238, 215)] 


    overlay = img.copy()
    # output = img.copy()
    # Plots one bounding box on image img
    tl = line_thickness or round(0.002 * (img.shape[0] + img.shape[1]) / 2) + 1  # line/font thickness

    color = list(att_colors[color_index])
    c1, c2 = (int(ref[0]), int(ref[1])), (int(vertex[0]), int(vertex[1]))
    cv2.line(overlay, c1, c2, (alpha*float(color[0])/255,alpha*float(color[1])/255,alpha*float(color[2])/255) , thickness=tl, lineType=cv2.LINE_AA)
    cv2.addWeighted(overlay, alpha, img, 1 - alpha, 0, img)



def cam2pixel(cam_coord, f, c):
    x = cam_coord[:, 0] / (cam_coord[:, 2]) * f[0] + c[0]
    y = cam_coord[:, 1] / (cam_coord[:, 2]) * f[1] + c[1]
    z = cam_coord[:, 2]
    img_coord = np.concatenate((x[:,None], y[:,None], z[:,None]),1)
    return img_coord


class Renderer(object):
    """
    Render mesh using OpenDR for visualization.
    """

    def __init__(self, width=800, height=600, near=0.5, far=1000, faces=None):
        self.colors = {'hand': [.9, .9, .9], 'pink': [.9, .7, .7], 'light_blue': [0.65098039, 0.74117647, 0.85882353] }
        self.width = width
        self.height = height
        self.faces = faces
        self.renderer = ColoredRenderer()

    def render(self, vertices, faces=None, img=None,
               camera_t=np.zeros([3], dtype=np.float32),
               camera_rot=np.zeros([3], dtype=np.float32),
               camera_center=None,
               use_bg=False,
               bg_color=(0.0, 0.0, 0.0),
               body_color=None,
               focal_length=5000,
               disp_text=False,
               gt_keyp=None,
               pred_keyp=None,
               **kwargs):
        if img is not None:
            height, width = img.shape[:2]
        else:
            height, width = self.height, self.width

        if faces is None:
            faces = self.faces

        if camera_center is None:
            camera_center = np.array([width * 0.5,
                                      height * 0.5])

        self.renderer.camera = ProjectPoints(rt=camera_rot,
                                             t=camera_t,
                                             f=focal_length * np.ones(2),
                                             c=camera_center,
                                             k=np.zeros(5))
        dist = np.abs(self.renderer.camera.t.r[2] -
                      np.mean(vertices, axis=0)[2])
        far = dist + 20

        self.renderer.frustum = {'near': 1.0, 'far': far,
                                 'width': width,
                                 'height': height}

        if img is not None:
            if use_bg:
                self.renderer.background_image = img
            else:
                self.renderer.background_image = np.ones_like(
                    img) * np.array(bg_color)

        if body_color is None:
            color = self.colors['light_blue']
        else:
            color = self.colors[body_color]

        if isinstance(self.renderer, TexturedRenderer):
            color = [1.,1.,1.]

        self.renderer.set(v=vertices, f=faces,
                          vc=color, bgcolor=np.ones(3))
        albedo = self.renderer.vc
        # Construct Back Light (on back right corner)
        yrot = np.radians(120)

        self.renderer.vc = LambertianPointLight(
            f=self.renderer.f,
            v=self.renderer.v,
            num_verts=self.renderer.v.shape[0],
            light_pos=rotateY(np.array([-200, -100, -100]), yrot),
            vc=albedo,
            light_color=np.array([1, 1, 1]))

        # Construct Left Light
        self.renderer.vc += LambertianPointLight(
            f=self.renderer.f,
            v=self.renderer.v,
            num_verts=self.renderer.v.shape[0],
            light_pos=rotateY(np.array([800, 10, 300]), yrot),
            vc=albedo,
            light_color=np.array([1, 1, 1]))

        #  Construct Right Light
        self.renderer.vc += LambertianPointLight(
            f=self.renderer.f,
            v=self.renderer.v,
            num_verts=self.renderer.v.shape[0],
            light_pos=rotateY(np.array([-500, 500, 1000]), yrot),
            vc=albedo,
            light_color=np.array([.7, .7, .7]))

        return self.renderer.r


    def render_vertex_color(self, vertices, faces=None, img=None,
               camera_t=np.zeros([3], dtype=np.float32),
               camera_rot=np.zeros([3], dtype=np.float32),
               camera_center=None,
               use_bg=False,
               bg_color=(0.0, 0.0, 0.0),
               vertex_color=None,
               focal_length=5000,
               disp_text=False,
               gt_keyp=None,
               pred_keyp=None,
               **kwargs):
        if img is not None:
            height, width = img.shape[:2]
        else:
            height, width = self.height, self.width

        if faces is None:
            faces = self.faces

        if camera_center is None:
            camera_center = np.array([width * 0.5,
                                      height * 0.5])

        self.renderer.camera = ProjectPoints(rt=camera_rot,
                                             t=camera_t,
                                             f=focal_length * np.ones(2),
                                             c=camera_center,
                                             k=np.zeros(5))
        dist = np.abs(self.renderer.camera.t.r[2] -
                      np.mean(vertices, axis=0)[2])
        far = dist + 20

        self.renderer.frustum = {'near': 1.0, 'far': far,
                                 'width': width,
                                 'height': height}

        if img is not None:
            if use_bg:
                self.renderer.background_image = img
            else:
                self.renderer.background_image = np.ones_like(
                    img) * np.array(bg_color)

        if vertex_color is None:
            vertex_color = self.colors['light_blue']


        self.renderer.set(v=vertices, f=faces,
                          vc=vertex_color, bgcolor=np.ones(3))
        albedo = self.renderer.vc
        # Construct Back Light (on back right corner)
        yrot = np.radians(120)

        self.renderer.vc = LambertianPointLight(
            f=self.renderer.f,
            v=self.renderer.v,
            num_verts=self.renderer.v.shape[0],
            light_pos=rotateY(np.array([-200, -100, -100]), yrot),
            vc=albedo,
            light_color=np.array([1, 1, 1]))

        # Construct Left Light
        self.renderer.vc += LambertianPointLight(
            f=self.renderer.f,
            v=self.renderer.v,
            num_verts=self.renderer.v.shape[0],
            light_pos=rotateY(np.array([800, 10, 300]), yrot),
            vc=albedo,
            light_color=np.array([1, 1, 1]))

        #  Construct Right Light
        self.renderer.vc += LambertianPointLight(
            f=self.renderer.f,
            v=self.renderer.v,
            num_verts=self.renderer.v.shape[0],
            light_pos=rotateY(np.array([-500, 500, 1000]), yrot),
            vc=albedo,
            light_color=np.array([.7, .7, .7]))

        return self.renderer.r


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.comfy.sampling import SamplerEnum
from nodetool.nodes.comfy.sampling import SchedulerEnum


class KSampler(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(
        default=UNet(type="comfy.unet"), description="The model to use."
    )
    seed: int | GraphNode | tuple[GraphNode, str] = Field(
        default=0, description="The seed to use."
    )
    steps: int | GraphNode | tuple[GraphNode, str] = Field(
        default=20, description="The number of steps to use."
    )
    cfg: float | GraphNode | tuple[GraphNode, str] = Field(
        default=8.0, description="The cfg to use."
    )
    sampler_name: SamplerEnum | GraphNode | tuple[GraphNode, str] = Field(
        default=SamplerEnum("ddim"), description="The sampler to use."
    )
    scheduler: SchedulerEnum | GraphNode | tuple[GraphNode, str] = Field(
        default=SchedulerEnum("normal"), description="The scheduler to use."
    )
    positive: Conditioning | GraphNode | tuple[GraphNode, str] = Field(
        default=Conditioning(type="comfy.conditioning"),
        description="The positive conditioning to use.",
    )
    negative: Conditioning | GraphNode | tuple[GraphNode, str] = Field(
        default=Conditioning(type="comfy.conditioning"),
        description="The negative conditioning to use.",
    )
    latent_image: Latent | GraphNode | tuple[GraphNode, str] = Field(
        default=Latent(type="comfy.latent"), description="The latent image to use."
    )
    denoise: float | GraphNode | tuple[GraphNode, str] = Field(
        default=1.0, description="The denoise to use."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.sampling.KSampler"


from nodetool.common.comfy_node import EnableDisable
from nodetool.nodes.comfy.sampling import SamplerEnum
from nodetool.nodes.comfy.sampling import SchedulerEnum
from nodetool.common.comfy_node import EnableDisable


class KSamplerAdvanced(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(
        default=UNet(type="comfy.unet"), description="The model to use."
    )
    add_noise: EnableDisable | GraphNode | tuple[GraphNode, str] = Field(
        default=EnableDisable("enable"), description="Enable or disable noise addition."
    )
    noise_seed: int | GraphNode | tuple[GraphNode, str] = Field(
        default=0, description="The seed for noise generation."
    )
    steps: int | GraphNode | tuple[GraphNode, str] = Field(
        default=20, description="The number of steps to use during sampling."
    )
    cfg: float | GraphNode | tuple[GraphNode, str] = Field(
        default=8.0, description="The configuration value for the sampler."
    )
    sampler_name: SamplerEnum | GraphNode | tuple[GraphNode, str] = Field(
        default=SamplerEnum("ddim"), description="The name of the sampler to use."
    )
    scheduler: SchedulerEnum | GraphNode | tuple[GraphNode, str] = Field(
        default=SchedulerEnum("normal"), description="The scheduler to use."
    )
    positive: Conditioning | GraphNode | tuple[GraphNode, str] = Field(
        default=Conditioning(type="comfy.conditioning"),
        description="The positive conditioning influence.",
    )
    negative: Conditioning | GraphNode | tuple[GraphNode, str] = Field(
        default=Conditioning(type="comfy.conditioning"),
        description="The negative conditioning influence.",
    )
    latent_image: Latent | GraphNode | tuple[GraphNode, str] = Field(
        default=Latent(type="comfy.latent"), description="The starting latent image."
    )
    start_at_step: int | GraphNode | tuple[GraphNode, str] = Field(
        default=0, description="The step at which to start the sampling process."
    )
    end_at_step: int | GraphNode | tuple[GraphNode, str] = Field(
        default=10000, description="The step at which to end the sampling process."
    )
    return_with_leftover_noise: EnableDisable | GraphNode | tuple[GraphNode, str] = (
        Field(
            default=EnableDisable("disable"),
            description="Whether to return with leftover noise or not.",
        )
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.sampling.KSamplerAdvanced"



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.replicate.image.analyze import Task

class Blip(GraphNode):
    task: Task | GraphNode | tuple[GraphNode, str] = Field(default='image_captioning', description='Choose a task.')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    caption: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Type caption for the input image for image text matching task.')
    question: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Type question for the input image for visual question answering task.')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.Blip"



class Blip2(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image to query or caption')
    caption: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Select if you want to generate image captions instead of asking questions')
    context: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Optional - previous questions and answers to be used as context for answering current question')
    question: str | GraphNode | tuple[GraphNode, str] = Field(default='What is this a picture of?', description='Question to ask about this image. Leave blank for captioning')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Temperature for use with nucleus sampling')
    use_nucleus_sampling: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Toggles the model using nucleus sampling to generate responses')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.Blip2"


from nodetool.nodes.replicate.image.analyze import Mode
from nodetool.nodes.replicate.image.analyze import Clip_model_name

class ClipInterrogator(GraphNode):
    mode: Mode | GraphNode | tuple[GraphNode, str] = Field(default='best', description='Prompt mode (best takes 10-20 seconds, fast takes 1-2 seconds).')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    clip_model_name: Clip_model_name | GraphNode | tuple[GraphNode, str] = Field(default='ViT-L-14/openai', description='Choose ViT-L for Stable Diffusion 1, ViT-H for Stable Diffusion 2, or ViT-bigG for Stable Diffusion XL.')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.ClipInterrogator"



class Llava13b(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='When decoding text, samples from the top p percentage of most likely tokens; lower to ignore less likely tokens')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Prompt to use for text generation')
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Maximum number of tokens to generate. A word is generally 2-3 tokens')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Adjusts randomness of outputs, greater than 1 is random and 0 is deterministic')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.Llava13b"



class Llava34B(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='When decoding text, samples from the top p percentage of most likely tokens; lower to ignore less likely tokens')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Prompt to use for text generation')
    history: list | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='List of earlier chat messages, alternating roles, starting with user input. Include <image> to specify which message to attach the image to.')
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=1024, description='Maximum number of tokens to generate. A word is generally 2-3 tokens')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Adjusts randomness of outputs, greater than 1 is random and 0 is deterministic')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.Llava34B"



class MiniGPT4(GraphNode):
    image: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Image to discuss')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=0.9, description='Sample from the top p percent most likely tokens')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Prompt for mini-gpt4 regarding input image')
    num_beams: int | GraphNode | tuple[GraphNode, str] = Field(default=3, description='Number of beams for beam search decoding')
    max_length: int | GraphNode | tuple[GraphNode, str] = Field(default=4000, description='Total length of prompt and output in tokens')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Temperature for generating tokens, lower = more predictable results')
    max_new_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=3000, description='Maximum number of new tokens to generate')
    repetition_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Penalty for repeated words in generated text; 1 is no penalty, values greater than 1 discourage repetition, less than 1 encourage it.')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.MiniGPT4"



class Moondream2(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='Describe this image', description='Input prompt')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.Moondream2"



class NSFWImageDetection(GraphNode):
    image: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Input image')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.NSFWImageDetection"


from nodetool.nodes.replicate.image.analyze import Mode

class SDXLClipInterrogator(GraphNode):
    mode: Mode | GraphNode | tuple[GraphNode, str] = Field(default='best', description='Prompt Mode: fast takes 1-2 seconds, best takes 15-25 seconds.')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    @classmethod
    def get_node_type(cls): return "replicate.image.analyze.SDXLClipInterrogator"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class LatentBatch(GraphNode):
    samples1: Latent | GraphNode | tuple[GraphNode, str] = Field(default=Latent(type='comfy.latent'), description='The first set of latent samples for the batch process.')
    samples2: Latent | GraphNode | tuple[GraphNode, str] = Field(default=Latent(type='comfy.latent'), description='The second set of latent samples for the batch process.')
    @classmethod
    def get_node_type(cls): return "comfy.latent.batch.LatentBatch"



class LatentFromBatch(GraphNode):
    samples: Latent | GraphNode | tuple[GraphNode, str] = Field(default=Latent(type='comfy.latent'), description='The batch of latent samples.')
    batch_index: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The index of the sample in the batch.')
    length: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='The length of latent samples to extract.')
    @classmethod
    def get_node_type(cls): return "comfy.latent.batch.LatentFromBatch"



class RepeatLatentBatch(GraphNode):
    samples: Latent | GraphNode | tuple[GraphNode, str] = Field(default=Latent(type='comfy.latent'), description='The latent samples to repeat.')
    amount: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='The amount of times to repeat each sample.')
    @classmethod
    def get_node_type(cls): return "comfy.latent.batch.RepeatLatentBatch"





from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class Language(str, Enum):
    AF = "af"
    AM = "am"
    AR = "ar"
    AS = "as"
    AZ = "az"
    BA = "ba"
    BE = "be"
    BG = "bg"
    BN = "bn"
    BO = "bo"
    BR = "br"
    BS = "bs"
    CA = "ca"
    CS = "cs"
    CY = "cy"
    DA = "da"
    DE = "de"
    EL = "el"
    EN = "en"
    ES = "es"
    ET = "et"
    EU = "eu"
    FA = "fa"
    FI = "fi"
    FO = "fo"
    FR = "fr"
    GL = "gl"
    GU = "gu"
    HA = "ha"
    HAW = "haw"
    HE = "he"
    HI = "hi"
    HR = "hr"
    HT = "ht"
    HU = "hu"
    HY = "hy"
    ID = "id"
    IS = "is"
    IT = "it"
    JA = "ja"
    JW = "jw"
    KA = "ka"
    KK = "kk"
    KM = "km"
    KN = "kn"
    KO = "ko"
    LA = "la"
    LB = "lb"
    LN = "ln"
    LO = "lo"
    LT = "lt"
    LV = "lv"
    MG = "mg"
    MI = "mi"
    MK = "mk"
    ML = "ml"
    MN = "mn"
    MR = "mr"
    MS = "ms"
    MT = "mt"
    MY = "my"
    NE = "ne"
    NL = "nl"
    NN = "nn"
    NO = "no"
    OC = "oc"
    PA = "pa"
    PL = "pl"
    PS = "ps"
    PT = "pt"
    RO = "ro"
    RU = "ru"
    SA = "sa"
    SD = "sd"
    SI = "si"
    SK = "sk"
    SL = "sl"
    SN = "sn"
    SO = "so"
    SQ = "sq"
    SR = "sr"
    SU = "su"
    SV = "sv"
    SW = "sw"
    TA = "ta"
    TE = "te"
    TG = "tg"
    TH = "th"
    TK = "tk"
    TL = "tl"
    TR = "tr"
    TT = "tt"
    UK = "uk"
    UR = "ur"
    UZ = "uz"
    VI = "vi"
    YI = "yi"
    YO = "yo"
    YUE = "yue"
    ZH = "zh"
    AFRIKAANS = "Afrikaans"
    ALBANIAN = "Albanian"
    AMHARIC = "Amharic"
    ARABIC = "Arabic"
    ARMENIAN = "Armenian"
    ASSAMESE = "Assamese"
    AZERBAIJANI = "Azerbaijani"
    BASHKIR = "Bashkir"
    BASQUE = "Basque"
    BELARUSIAN = "Belarusian"
    BENGALI = "Bengali"
    BOSNIAN = "Bosnian"
    BRETON = "Breton"
    BULGARIAN = "Bulgarian"
    BURMESE = "Burmese"
    CANTONESE = "Cantonese"
    CASTILIAN = "Castilian"
    CATALAN = "Catalan"
    CHINESE = "Chinese"
    CROATIAN = "Croatian"
    CZECH = "Czech"
    DANISH = "Danish"
    DUTCH = "Dutch"
    ENGLISH = "English"
    ESTONIAN = "Estonian"
    FAROESE = "Faroese"
    FINNISH = "Finnish"
    FLEMISH = "Flemish"
    FRENCH = "French"
    GALICIAN = "Galician"
    GEORGIAN = "Georgian"
    GERMAN = "German"
    GREEK = "Greek"
    GUJARATI = "Gujarati"
    HAITIAN = "Haitian"
    HAITIAN_CREOLE = "Haitian Creole"
    HAUSA = "Hausa"
    HAWAIIAN = "Hawaiian"
    HEBREW = "Hebrew"
    HINDI = "Hindi"
    HUNGARIAN = "Hungarian"
    ICELANDIC = "Icelandic"
    INDONESIAN = "Indonesian"
    ITALIAN = "Italian"
    JAPANESE = "Japanese"
    JAVANESE = "Javanese"
    KANNADA = "Kannada"
    KAZAKH = "Kazakh"
    KHMER = "Khmer"
    KOREAN = "Korean"
    LAO = "Lao"
    LATIN = "Latin"
    LATVIAN = "Latvian"
    LETZEBURGESCH = "Letzeburgesch"
    LINGALA = "Lingala"
    LITHUANIAN = "Lithuanian"
    LUXEMBOURGISH = "Luxembourgish"
    MACEDONIAN = "Macedonian"
    MALAGASY = "Malagasy"
    MALAY = "Malay"
    MALAYALAM = "Malayalam"
    MALTESE = "Maltese"
    MANDARIN = "Mandarin"
    MAORI = "Maori"
    MARATHI = "Marathi"
    MOLDAVIAN = "Moldavian"
    MOLDOVAN = "Moldovan"
    MONGOLIAN = "Mongolian"
    MYANMAR = "Myanmar"
    NEPALI = "Nepali"
    NORWEGIAN = "Norwegian"
    NYNORSK = "Nynorsk"
    OCCITAN = "Occitan"
    PANJABI = "Panjabi"
    PASHTO = "Pashto"
    PERSIAN = "Persian"
    POLISH = "Polish"
    PORTUGUESE = "Portuguese"
    PUNJABI = "Punjabi"
    PUSHTO = "Pushto"
    ROMANIAN = "Romanian"
    RUSSIAN = "Russian"
    SANSKRIT = "Sanskrit"
    SERBIAN = "Serbian"
    SHONA = "Shona"
    SINDHI = "Sindhi"
    SINHALA = "Sinhala"
    SINHALESE = "Sinhalese"
    SLOVAK = "Slovak"
    SLOVENIAN = "Slovenian"
    SOMALI = "Somali"
    SPANISH = "Spanish"
    SUNDANESE = "Sundanese"
    SWAHILI = "Swahili"
    SWEDISH = "Swedish"
    TAGALOG = "Tagalog"
    TAJIK = "Tajik"
    TAMIL = "Tamil"
    TATAR = "Tatar"
    TELUGU = "Telugu"
    THAI = "Thai"
    TIBETAN = "Tibetan"
    TURKISH = "Turkish"
    TURKMEN = "Turkmen"
    UKRAINIAN = "Ukrainian"
    URDU = "Urdu"
    UZBEK = "Uzbek"
    VALENCIAN = "Valencian"
    VIETNAMESE = "Vietnamese"
    WELSH = "Welsh"
    YIDDISH = "Yiddish"
    YORUBA = "Yoruba"


class Transcription(str, Enum):
    PLAIN_TEXT = "plain text"
    SRT = "srt"
    VTT = "vtt"


class Whisper(ReplicateNode):
    """Convert speech in audio to text"""

    @classmethod
    def replicate_model_id(cls):
        return "openai/whisper:4d50797290df275329f202e48c76360b3f22b08d28c196cbc54600319435f8d2"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 (High-memory) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/a9253a5b-aa79-4df3-8aa1-09a6d044fdd6/Screen_Shot_2022-09-28_at_16.58.5.png",
            "created_at": "2022-09-22T10:01:42.119733Z",
            "description": "Convert speech in audio to text",
            "github_url": "https://github.com/replicate/cog-whisper",
            "license_url": "https://github.com/openai/whisper/blob/main/LICENSE",
            "name": "whisper",
            "owner": "openai",
            "paper_url": "https://cdn.openai.com/papers/whisper.pdf",
            "run_count": 6339499,
            "url": "https://replicate.com/openai/whisper",
            "visibility": "public",
            "hardware": "Nvidia T4 (High-memory) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    audio: AudioRef = Field(default=AudioRef(), description="Audio file")
    model: str = Field(
        title="Model",
        description="This version only supports Whisper-large-v3.",
        default="large-v3",
    )
    language: Language = Field(
        description="language spoken in the audio, specify None to perform language detection",
        default=None,
    )
    patience: float | None = Field(
        title="Patience",
        description="optional patience value to use in beam decoding, as in https://arxiv.org/abs/2204.05424, the default (1.0) is equivalent to conventional beam search",
        default=None,
    )
    translate: bool = Field(
        title="Translate",
        description="Translate the text to English when set to True",
        default=False,
    )
    temperature: float = Field(
        title="Temperature", description="temperature to use for sampling", default=0
    )
    transcription: Transcription = Field(
        description="Choose the format for the transcription", default="plain text"
    )
    initial_prompt: str | None = Field(
        title="Initial Prompt",
        description="optional text to provide as a prompt for the first window.",
        default=None,
    )
    suppress_tokens: str = Field(
        title="Suppress Tokens",
        description="comma-separated list of token ids to suppress during sampling; '-1' will suppress most special characters except common punctuations",
        default="-1",
    )
    logprob_threshold: float = Field(
        title="Logprob Threshold",
        description="if the average log probability is lower than this value, treat the decoding as failed",
        default=-1,
    )
    no_speech_threshold: float = Field(
        title="No Speech Threshold",
        description="if the probability of the <|nospeech|> token is higher than this value AND the decoding has failed due to `logprob_threshold`, consider the segment as silence",
        default=0.6,
    )
    condition_on_previous_text: bool = Field(
        title="Condition On Previous Text",
        description="if True, provide the previous output of the model as a prompt for the next window; disabling may make the text inconsistent across windows, but the model becomes less prone to getting stuck in a failure loop",
        default=True,
    )
    compression_ratio_threshold: float = Field(
        title="Compression Ratio Threshold",
        description="if the gzip compression ratio is higher than this value, treat the decoding as failed",
        default=2.4,
    )
    temperature_increment_on_fallback: float = Field(
        title="Temperature Increment On Fallback",
        description="temperature to increase when falling back when the decoding fails to meet either of the thresholds below",
        default=0.2,
    )


class Task(str, Enum):
    TRANSCRIBE = "transcribe"
    TRANSLATE = "translate"


class Timestamp(str, Enum):
    CHUNK = "chunk"
    WORD = "word"


class IncrediblyFastWhisper(ReplicateNode):
    """whisper-large-v3, incredibly fast, powered by Hugging Face Transformers! 🤗"""

    @classmethod
    def replicate_model_id(cls):
        return "vaibhavs10/incredibly-fast-whisper:3ab86df6c8f54c11309d4d1f930ac292bad43ace52d10c80d87eb258b3c9f79c"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 (Large) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/a59dc78f-523f-4977-9856-e84479aec79b/insanely-fast-whisper-img.png",
            "created_at": "2023-11-13T13:28:53.689979Z",
            "description": "whisper-large-v3, incredibly fast, powered by Hugging Face Transformers! 🤗",
            "github_url": "https://github.com/chenxwh/insanely-fast-whisper",
            "license_url": "https://github.com/Vaibhavs10/insanely-fast-whisper/blob/main/LICENSE",
            "name": "incredibly-fast-whisper",
            "owner": "vaibhavs10",
            "paper_url": None,
            "run_count": 300039,
            "url": "https://replicate.com/vaibhavs10/incredibly-fast-whisper",
            "visibility": "public",
            "hardware": "Nvidia A40 (Large) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    task: Task = Field(
        description="Task to perform: transcribe or translate to another language.",
        default="transcribe",
    )
    audio: AudioRef = Field(default=AudioRef(), description="Audio file")
    hf_token: str | None = Field(
        title="Hf Token",
        description="Provide a hf.co/settings/token for Pyannote.audio to diarise the audio clips. You need to agree to the terms in 'https://huggingface.co/pyannote/speaker-diarization-3.1' and 'https://huggingface.co/pyannote/segmentation-3.0' first.",
        default=None,
    )
    language: Language = Field(
        description="Language spoken in the audio, specify 'None' to perform language detection.",
        default="None",
    )
    timestamp: Timestamp = Field(
        description="Whisper supports both chunked as well as word level timestamps.",
        default="chunk",
    )
    batch_size: int = Field(
        title="Batch Size",
        description="Number of parallel batches you want to compute. Reduce if you face OOMs.",
        default=24,
    )
    diarise_audio: bool = Field(
        title="Diarise Audio",
        description="Use Pyannote.audio to diarise the audio clips. You will need to provide hf_token below too.",
        default=False,
    )



import numpy as np
import cv2
import os
import cv2
from .dw_onnx.cv_ox_det import inference_detector as inference_onnx_yolox
from .dw_onnx.cv_ox_yolo_nas import inference_detector as inference_onnx_yolo_nas
from .dw_onnx.cv_ox_pose import inference_pose as inference_onnx_pose

from .dw_torchscript.jit_det import inference_detector as inference_jit_yolox
from .dw_torchscript.jit_pose import inference_pose as inference_jit_pose
from typing import List, Optional
from .types import PoseResult, BodyResult, Keypoint
from timeit import default_timer
from controlnet_aux.dwpose.util import guess_onnx_input_shape_dtype, get_ort_providers, get_model_type, is_model_torchscript
import json
import torch
import torch.utils.benchmark.utils.timer as torch_timer

def drawBetweenKeypoints(pose_img, keypoints, indexes, color, scaleFactor):
    ind0 = indexes[0] - 1
    ind1 = indexes[1] - 1
    
    point1 = (keypoints[ind0][0], keypoints[ind0][1])
    point2 = (keypoints[ind1][0], keypoints[ind1][1])

    thickness = int(5 // scaleFactor)


    cv2.line(pose_img, (int(point1[0]), int(point1[1])), (int(point2[0]), int(point2[1])), color, thickness)


def drawBetweenKeypointsList(pose_img, keypoints, keypointPairsList, colorsList, scaleFactor):
    for ind, keypointPair in enumerate(keypointPairsList):
        drawBetweenKeypoints(pose_img, keypoints, keypointPair, colorsList[ind], scaleFactor)

def drawBetweenSetofKeypointLists(pose_img, keypoints_set, keypointPairsList, colorsList, scaleFactor):
    for keypoints in keypoints_set:
        drawBetweenKeypointsList(pose_img, keypoints, keypointPairsList, colorsList, scaleFactor)


def padImg(img, size, blackBorder=True):
    left, right, top, bottom = 0, 0, 0, 0

    # pad x
    if img.shape[1] < size[1]:
        sidePadding = int((size[1] - img.shape[1]) // 2)
        left = sidePadding
        right = sidePadding

        # pad extra on right if padding needed is an odd number
        if img.shape[1] % 2 == 1:
            right += 1

    # pad y
    if img.shape[0] < size[0]:
        topBottomPadding = int((size[0] - img.shape[0]) // 2)
        top = topBottomPadding
        bottom = topBottomPadding
        
        # pad extra on bottom if padding needed is an odd number
        if img.shape[0] % 2 == 1:
            bottom += 1

    if blackBorder:
        paddedImg = cv2.copyMakeBorder(src=img, top=top, bottom=bottom, left=left, right=right, borderType=cv2.BORDER_CONSTANT, value=(0,0,0))
    else:
        paddedImg = cv2.copyMakeBorder(src=img, top=top, bottom=bottom, left=left, right=right, borderType=cv2.BORDER_REPLICATE)

    return paddedImg

def smartCrop(img, size, center):

    width = img.shape[1]
    height = img.shape[0]
    xSize = size[1]
    ySize = size[0]
    xCenter = center[0]
    yCenter = center[1]

    if img.shape[0] > size[0] or img.shape[1] > size[1]:


        leftMargin = xCenter - xSize//2
        rightMargin = xCenter + xSize//2
        upMargin = yCenter - ySize//2
        downMargin = yCenter + ySize//2


        if(leftMargin < 0):
            xCenter += (-leftMargin)
        if(rightMargin > width):
            xCenter -= (rightMargin - width)

        if(upMargin < 0):
            yCenter -= -upMargin
        if(downMargin > height):
            yCenter -= (downMargin - height)


        img = cv2.getRectSubPix(img, size, (xCenter, yCenter))



    return img



def calculateScaleFactor(img, size, poseSpanX, poseSpanY):

    poseSpanX = max(poseSpanX, size[0])

    scaleFactorX = 1


    if poseSpanX > size[0]:
        scaleFactorX = size[0] / poseSpanX

    scaleFactorY = 1
    if poseSpanY > size[1]:
        scaleFactorY = size[1] / poseSpanY

    scaleFactor = min(scaleFactorX, scaleFactorY)


    return scaleFactor



def scaleImg(img, size, poseSpanX, poseSpanY, scaleFactor):
    scaledImg = img

    scaledImg = cv2.resize(img, (0, 0), fx=scaleFactor, fy=scaleFactor) 

    return scaledImg, scaleFactor

class AnimalPoseImage:
    def __init__(self, det_model_path: Optional[str] = None, pose_model_path: Optional[str] = None, torchscript_device="cuda"):
        self.det_filename = det_model_path and os.path.basename(det_model_path)
        self.pose_filename = pose_model_path and os.path.basename(pose_model_path)
        self.det, self.pose = None, None
        # return type: None ort cv2 torchscript
        self.det_model_type = get_model_type("AnimalPose",self.det_filename)
        self.pose_model_type = get_model_type("AnimalPose",self.pose_filename)
        # Always loads to CPU to avoid building OpenCV.
        cv2_device = 'cpu'
        cv2_backend = cv2.dnn.DNN_BACKEND_OPENCV if cv2_device == 'cpu' else cv2.dnn.DNN_BACKEND_CUDA
        # You need to manually build OpenCV through cmake to work with your GPU.
        cv2_providers = cv2.dnn.DNN_TARGET_CPU if cv2_device == 'cpu' else cv2.dnn.DNN_TARGET_CUDA
        ort_providers = get_ort_providers()

        if self.det_model_type is None:
            pass
        elif self.det_model_type == "ort":
            try:
                import onnxruntime as ort
                self.det = ort.InferenceSession(det_model_path, providers=ort_providers)
            except:
                print(f"Failed to load onnxruntime with {self.det.get_providers()}.\nPlease change EP_list in the config.yaml and restart ComfyUI")
                self.det = ort.InferenceSession(det_model_path, providers=["CPUExecutionProvider"])
        elif self.det_model_type == "cv2":
            try:
                self.det = cv2.dnn.readNetFromONNX(det_model_path)
                self.det.setPreferableBackend(cv2_backend)
                self.det.setPreferableTarget(cv2_providers)
            except:
                print("TopK operators may not work on your OpenCV, try use onnxruntime with CPUExecutionProvider")
                try:
                    import onnxruntime as ort
                    self.det = ort.InferenceSession(det_model_path, providers=["CPUExecutionProvider"])
                except:
                    print(f"Failed to load {det_model_path}, you can use other models instead")
        else:
            self.det = torch.jit.load(det_model_path)
            self.det.to(torchscript_device)

        if self.pose_model_type is None:
            pass
        elif self.pose_model_type == "ort":
            try:
                import onnxruntime as ort
                self.pose = ort.InferenceSession(pose_model_path, providers=ort_providers)
            except:
                print(f"Failed to load onnxruntime with {self.pose.get_providers()}.\nPlease change EP_list in the config.yaml and restart ComfyUI")
                self.pose = ort.InferenceSession(pose_model_path, providers=["CPUExecutionProvider"])
        elif self.pose_model_type == "cv2":
            self.pose = cv2.dnn.readNetFromONNX(pose_model_path)
            self.pose.setPreferableBackend(cv2_backend)
            self.pose.setPreferableTarget(cv2_providers)
        else:
            self.pose = torch.jit.load(pose_model_path)
            self.pose.to(torchscript_device)

        if self.pose_filename is not None:
            self.pose_input_size, _ = guess_onnx_input_shape_dtype(self.pose_filename)
    
    def __call__(self, oriImg) -> Optional[np.ndarray]:
        detect_classes = list(range(14, 23 + 1)) #https://github.com/ultralytics/ultralytics/blob/main/ultralytics/cfg/datasets/coco.yaml

        if is_model_torchscript(self.det):
            det_start = torch_timer.timer()
            det_result = inference_jit_yolox(self.det, oriImg, detect_classes=detect_classes)
            print(f"AnimalPose: Bbox {((torch_timer.timer() - det_start) * 1000):.2f}ms")
        else:
            det_start = default_timer()
            det_onnx_dtype = np.float32 if "yolox" in self.det_filename else np.uint8
            if "yolox" in self.det_filename:
                det_result = inference_onnx_yolox(self.det, oriImg, detect_classes=detect_classes, dtype=det_onnx_dtype)
            else:
                #FP16 and INT8 YOLO NAS accept uint8 input
                det_result = inference_onnx_yolo_nas(self.det, oriImg, detect_classes=detect_classes, dtype=det_onnx_dtype)
            print(f"AnimalPose: Bbox {((default_timer() - det_start) * 1000):.2f}ms")
        if (det_result is None) or (det_result.shape[0] == 0):
            openpose_dict = {
                'version': 'ap10k',
                'animals': [],
                'canvas_height': oriImg.shape[0],
                'canvas_width': oriImg.shape[1]
            }
            return np.zeros_like(oriImg), openpose_dict
        
        if is_model_torchscript(self.pose):
            pose_start = torch_timer.timer()
            keypoint_sets, scores = inference_jit_pose(self.pose, det_result, oriImg, self.pose_input_size)
            print(f"AnimalPose: Pose {((torch_timer.timer() - pose_start) * 1000):.2f}ms on {det_result.shape[0]} animals\n")
        else:
            pose_start = default_timer()
            _, pose_onnx_dtype = guess_onnx_input_shape_dtype(self.pose_filename)
            keypoint_sets, scores = inference_onnx_pose(self.pose, det_result, oriImg, self.pose_input_size, dtype=pose_onnx_dtype)
            print(f"AnimalPose: Pose {((default_timer() - pose_start) * 1000):.2f}ms on {det_result.shape[0]} animals\n")

        animal_kps_scores = []
        pose_img = np.zeros((oriImg.shape[0], oriImg.shape[1], 3), dtype = np.uint8)
        for (idx, keypoints) in enumerate(keypoint_sets):
            # don't use keypoints that go outside the frame in calculations for the center
            interorKeypoints = keypoints[((keypoints[:,0] > 0) & (keypoints[:,0] < oriImg.shape[1])) & ((keypoints[:,1] > 0) & (keypoints[:,1] < oriImg.shape[0]))]

            xVals = interorKeypoints[:,0]
            yVals = interorKeypoints[:,1]

            minX = np.amin(xVals)
            minY = np.amin(yVals)
            maxX = np.amax(xVals)
            maxY = np.amax(yVals)

            poseSpanX = maxX - minX
            poseSpanY = maxY - minY

            # find mean center

            xSum = np.sum(xVals)
            ySum = np.sum(yVals)

            xCenter = xSum // xVals.shape[0]
            yCenter = ySum // yVals.shape[0]
            center_of_keypoints = (xCenter,yCenter)

            # order of the keypoints for AP10k and a standardized list of colors for limbs
            keypointPairsList = [(1,2), (2,3), (1,3), (3,4), (4,9), (9,10), (10,11), (4,6), (6,7), (7,8), (4,5), (5,15), (15,16), (16,17), (5,12), (12,13), (13,14)]
            colorsList = [(255,255,255), (100,255,100), (150,255,255), (100,50,255), (50,150,200), (0,255,255), (0,150,0), (0,0,255), (0,0,150), (255,50,255), (255,0,255), (255,0,0), (150,0,0), (255,255,100), (0,150,0), (255,255,0), (150,150,150)] # 16 colors needed

            drawBetweenKeypointsList(pose_img, keypoints, keypointPairsList, colorsList, scaleFactor=1.0)
            score = scores[idx, ..., None]
            score[score > 1.0] = 1.0
            score[score < 0.0] = 0.0
            animal_kps_scores.append(np.concatenate((keypoints, score), axis=-1))
        
        openpose_dict = {
            'version': 'ap10k',
            'animals': [keypoints.tolist() for keypoints in animal_kps_scores],
            'canvas_height': oriImg.shape[0],
            'canvas_width': oriImg.shape[1]
        }
        return pose_img, openpose_dict


#!/usr/bin/env python

from datetime import datetime
from fastapi import APIRouter, Depends, HTTPException
from nodetool.api.utils import current_user, User
from nodetool.api.types.chat import (
    TaskCreateRequest,
    TaskUpdateRequest,
    TaskList,
    Task,
)
from nodetool.common.environment import Environment
from nodetool.models.task import Task as TaskModel
from typing import Optional

log = Environment.get_logger()
router = APIRouter(prefix="/api/tasks", tags=["tasks"])


@router.get("/")
async def index(
    thread_id: str,
    user: User = Depends(current_user),
    cursor: Optional[str] = None,
    page_size: Optional[int] = None,
) -> TaskList:
    """
    Returns all tasks for the current user, optionally filtered by status.
    """
    if page_size is None:
        page_size = 100

    tasks, next_cursor = TaskModel.paginate(
        thread_id=thread_id, limit=page_size, start_key=cursor
    )

    # validate the tasks are from the current user
    tasks = [task for task in tasks if task.user_id == user.id]

    return TaskList(next=next_cursor, tasks=[Task.from_model(task) for task in tasks])


@router.get("/{id}")
async def get(id: str, user: User = Depends(current_user)) -> Task:
    """
    Returns the task with the given id.
    """
    task = TaskModel.find(user.id, id)
    if task is None:
        raise HTTPException(status_code=404, detail="Task not found")
    return Task.from_model(task)


@router.post("/")
async def create(req: TaskCreateRequest, user: User = Depends(current_user)) -> Task:
    """
    Creates a new task.
    """
    task = TaskModel.create(
        user_id=user.id,
        thread_id=req.thread_id,
        task_type=req.task_type,
        name=req.name,
        instructions=req.instructions,
        dependencies=req.dependencies,
        required_capabilities=req.required_capabilities,
        started_at=datetime.now(),
    )
    return Task.from_model(task)


@router.put("/{id}")
async def update(
    id: str, req: TaskUpdateRequest, user: User = Depends(current_user)
) -> Task:
    """
    Updates the task with the given id.
    """
    task = TaskModel.find(user.id, id)
    if task is None:
        raise HTTPException(status_code=404, detail="Task not found")

    if req.status:
        task.status = req.status
    if req.error:
        task.error = req.error
    if req.result:
        task.result = req.result
    if req.cost:
        task.cost = req.cost

    task.save()
    return Task.from_model(task)


@router.delete("/{id}")
async def delete(id: str, user: User = Depends(current_user)):
    """
    Deletes the task with the given id.
    """
    task = TaskModel.find(user.id, id)
    if task is None:
        raise HTTPException(status_code=404, detail="Task not found")
    task.delete()



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class Mode(str, Enum):
    SMALL = 'small'
    MEDIUM = 'medium'
    UPSCALED = 'upscaled'
    UPSCALED_AND_INTERPOLATED = 'upscaled-and-interpolated'
class Checkpoint(str, Enum):
    REALISTIC = 'realistic'
    ILLUSTRATED = 'illustrated'
    ANIME = 'anime'
    _3D = '3D'
    ANY = 'any'
class Aspect_ratio(str, Enum):
    _16_9 = '16:9'
    _4_3 = '4:3'
    _3_2 = '3:2'
    _1_1 = '1:1'
    _2_3 = '2:3'
    _3_4 = '3:4'
    _9_16 = '9:16'


class VideoMorpher(ReplicateNode):
    """Generate a video that morphs between subjects, with an optional style"""

    @classmethod
    def replicate_model_id(cls): return 'fofr/video-morpher:e70e975067d2b5dbe9e2d9022833d27230a1bdeb3f4af6fe6bb49a548a3039a7'
    @classmethod
    def get_hardware(cls): return 'Nvidia A40 (Large) GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/cd22e70a-5a93-463d-87ae-bb328d3a2573/ezgif-7-39b54a9237.gif', 'created_at': '2024-04-24T13:25:30.032622Z', 'description': 'Generate a video that morphs between subjects, with an optional style', 'github_url': 'https://github.com/fofr/cog-video-morpher', 'license_url': 'https://github.com/fofr/cog-video-morpher/blob/main/LICENSE', 'name': 'video-morpher', 'owner': 'fofr', 'paper_url': None, 'run_count': 6491, 'url': 'https://replicate.com/fofr/video-morpher', 'visibility': 'public', 'hardware': 'Nvidia A40 (Large) GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    mode: Mode = Field(description='Determines if you produce a quick experimental video or an upscaled interpolated one. (small ~20s, medium ~60s, upscaled ~2min, upscaled-and-interpolated ~4min)', default='medium')
    seed: int | None = Field(title='Seed', description='Set a seed for reproducibility. Random by default.', default=None)
    prompt: str = Field(title='Prompt', description='The prompt has a small effect, but most of the video is driven by the subject images', default='')
    checkpoint: Checkpoint = Field(description='The checkpoint to use for the model', default='realistic')
    style_image: ImageRef = Field(default=ImageRef(), description='Apply the style from this image to the whole video')
    aspect_ratio: Aspect_ratio = Field(description='The aspect ratio of the video', default='2:3')
    style_strength: float = Field(title='Style Strength', description='How strong the style is applied', ge=0.0, le=2.0, default=1)
    use_controlnet: bool = Field(title='Use Controlnet', description='Use geometric circles to guide the generation', default=True)
    negative_prompt: str = Field(title='Negative Prompt', description='What you do not want to see in the video', default='')
    subject_image_1: ImageRef = Field(default=ImageRef(), description='The first subject of the video')
    subject_image_2: ImageRef = Field(default=ImageRef(), description='The second subject of the video')
    subject_image_3: ImageRef = Field(default=ImageRef(), description='The third subject of the video')
    subject_image_4: ImageRef = Field(default=ImageRef(), description='The fourth subject of the video')

class Width(int, Enum):
    _256 = 256
    _320 = 320
    _384 = 384
    _448 = 448
    _512 = 512
    _576 = 576
    _640 = 640
    _672 = 672
    _704 = 704
    _768 = 768
    _832 = 832
    _896 = 896
    _960 = 960
    _1024 = 1024
class Height(int, Enum):
    _256 = 256
    _320 = 320
    _384 = 384
    _448 = 448
    _512 = 512
    _576 = 576
    _640 = 640
    _672 = 672
    _704 = 704
    _768 = 768
    _832 = 832
    _896 = 896
    _960 = 960
    _1024 = 1024
class Scheduler(str, Enum):
    DDIMSCHEDULER = 'DDIMScheduler'
    DPMSOLVERMULTISTEPSCHEDULER = 'DPMSolverMultistepScheduler'
    HEUNDISCRETESCHEDULER = 'HeunDiscreteScheduler'
    KARRASDPM = 'KarrasDPM'
    EULERANCESTRALDISCRETESCHEDULER = 'EulerAncestralDiscreteScheduler'
    EULERDISCRETESCHEDULER = 'EulerDiscreteScheduler'
    PNDMSCHEDULER = 'PNDMScheduler'


class HotshotXL(ReplicateNode):
    """😊 Hotshot-XL is an AI text-to-GIF model trained to work alongside Stable Diffusion XL"""

    @classmethod
    def replicate_model_id(cls): return 'lucataco/hotshot-xl:78b3a6257e16e4b241245d65c8b2b81ea2e1ff7ed4c55306b511509ddbfd327a'
    @classmethod
    def get_hardware(cls): return 'Nvidia A40 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/ed1749eb-87a6-45a9-bff3-cac89b7df97a/output.gif', 'created_at': '2023-10-05T04:09:21.646870Z', 'description': '😊 Hotshot-XL is an AI text-to-GIF model trained to work alongside Stable Diffusion XL', 'github_url': 'https://github.com/lucataco/cog-hotshot-xl', 'license_url': 'https://github.com/hotshotco/Hotshot-XL/blob/main/LICENSE', 'name': 'hotshot-xl', 'owner': 'lucataco', 'paper_url': 'https://huggingface.co/hotshotco/SDXL-512', 'run_count': 100582, 'url': 'https://replicate.com/lucataco/hotshot-xl', 'visibility': 'public', 'hardware': 'Nvidia A40 GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    mp4: bool = Field(title='Mp4', description='Save as mp4, False for GIF', default=False)
    seed: int | None = Field(title='Seed', description='Random seed. Leave blank to randomize the seed', default=None)
    steps: int = Field(title='Steps', description='Number of denoising steps', ge=1.0, le=500.0, default=30)
    width: Width = Field(description='Width of the output', default=672)
    height: Height = Field(description='Height of the output', default=384)
    prompt: str = Field(title='Prompt', description='Input prompt', default='a camel smoking a cigarette, hd, high quality')
    scheduler: Scheduler = Field(description='Select a Scheduler', default='EulerAncestralDiscreteScheduler')
    negative_prompt: str = Field(title='Negative Prompt', description='Negative prompt', default='blurry')

class Base_model(str, Enum):
    REALISTICVISIONV20_V20 = 'realisticVisionV20_v20'
    LYRIEL_V16 = 'lyriel_v16'
    MAJICMIXREALISTIC_V5PREVIEW = 'majicmixRealistic_v5Preview'
    RCNZCARTOON3D_V10 = 'rcnzCartoon3d_v10'
    TOONYOU_BETA3 = 'toonyou_beta3'
class Output_format(str, Enum):
    MP4 = 'mp4'
    GIF = 'gif'


class AnimateDiff(ReplicateNode):
    """🎨 AnimateDiff (w/ MotionLoRAs for Panning, Zooming, etc): Animate Your Personalized Text-to-Image Diffusion Models without Specific Tuning"""

    @classmethod
    def replicate_model_id(cls): return 'zsxkib/animate-diff:269a616c8b0c2bbc12fc15fd51bb202b11e94ff0f7786c026aa905305c4ed9fb'
    @classmethod
    def get_hardware(cls): return 'Nvidia A40 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/336d5c4e-4fd1-415a-a3f7-4a40fa5bdf2b/a_middle-aged_woman_utilizing__zo.gif', 'created_at': '2023-09-26T19:55:56.734880Z', 'description': '🎨 AnimateDiff (w/ MotionLoRAs for Panning, Zooming, etc): Animate Your Personalized Text-to-Image Diffusion Models without Specific Tuning', 'github_url': 'https://github.com/guoyww/AnimateDiff', 'license_url': 'https://github.com/guoyww/AnimateDiff/blob/main/LICENSE.txt', 'name': 'animate-diff', 'owner': 'zsxkib', 'paper_url': 'https://arxiv.org/abs/2307.04725', 'run_count': 39379, 'url': 'https://replicate.com/zsxkib/animate-diff', 'visibility': 'public', 'hardware': 'Nvidia A40 GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    seed: int = Field(title='Seed', description='Seed for different images and reproducibility. Use -1 to randomise seed', default=-1)
    steps: int = Field(title='Steps', description='Number of inference steps', ge=1.0, le=100.0, default=25)
    width: int = Field(title='Width', description='Width in pixels', default=512)
    frames: int = Field(title='Frames', description='Length of the video in frames (playback is at 8 fps e.g. 16 frames @ 8 fps is 2 seconds)', ge=1.0, le=32.0, default=16)
    height: int = Field(title='Height', description='Height in pixels', default=512)
    prompt: str = Field(title='Prompt', default='photo of vocano, rocks, storm weather, wind, lava waves, lightning, 8k uhd, dslr, soft lighting, high quality, film grain, Fujifilm XT3')
    base_model: Base_model = Field(description='Select a base model (DreamBooth checkpoint)', default='realisticVisionV20_v20')
    output_format: Output_format = Field(description="Output format of the video. Can be 'mp4' or 'gif'", default='mp4')
    guidance_scale: float = Field(title='Guidance Scale', description='Guidance Scale. How closely do we want to adhere to the prompt and its contents', ge=0.0, le=20.0, default=7.5)
    negative_prompt: str = Field(title='Negative Prompt', default='blur, haze, deformed iris, deformed pupils, semi-realistic, cgi, 3d, render, sketch, cartoon, drawing, anime, mutated hands and fingers, deformed, distorted, disfigured, poorly drawn, bad anatomy, wrong anatomy, extra limb, missing limb, floating limbs, disconnected limbs, mutation, mutated, ugly, disgusting, amputation')
    pan_up_motion_strength: float = Field(title='Pan Up Motion Strength', description='Strength of Pan Up Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    zoom_in_motion_strength: float = Field(title='Zoom In Motion Strength', description='Strength of Zoom In Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    pan_down_motion_strength: float = Field(title='Pan Down Motion Strength', description='Strength of Pan Down Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    pan_left_motion_strength: float = Field(title='Pan Left Motion Strength', description='Strength of Pan Left Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    zoom_out_motion_strength: float = Field(title='Zoom Out Motion Strength', description='Strength of Zoom Out Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    pan_right_motion_strength: float = Field(title='Pan Right Motion Strength', description='Strength of Pan Right Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    rolling_clockwise_motion_strength: float = Field(title='Rolling Clockwise Motion Strength', description='Strength of Rolling Clockwise Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)
    rolling_anticlockwise_motion_strength: float = Field(title='Rolling Anticlockwise Motion Strength', description='Strength of Rolling Anticlockwise Motion LoRA. 0 disables the LoRA', ge=0.0, le=1.0, default=0)



class Tooncrafter(ReplicateNode):
    """Create videos from illustrated input images"""

    @classmethod
    def replicate_model_id(cls): return 'fofr/tooncrafter:51bf654d60d307ab45c4ffe09546a3c9606f8f33861ab28f5bb0e43ad3fa40ed'
    @classmethod
    def get_hardware(cls): return 'Nvidia A100 (40GB) GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/e3adaa0a-c534-496c-81d1-281f9000fbb4/tooncrafter.gif', 'created_at': '2024-06-02T21:29:15.300292Z', 'description': 'Create videos from illustrated input images', 'github_url': 'https://github.com/fofr/cog-comfyui-tooncrafter', 'license_url': 'https://github.com/fofr/cog-comfyui-tooncrafter/blob/main/LICENSE', 'name': 'tooncrafter', 'owner': 'fofr', 'paper_url': 'https://doubiiu.github.io/projects/ToonCrafter/', 'run_count': 7390, 'url': 'https://replicate.com/fofr/tooncrafter', 'visibility': 'public', 'hardware': 'Nvidia A100 (40GB) GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    loop: bool = Field(title='Loop', description='Loop the video', default=False)
    seed: int | None = Field(title='Seed', description='Set a seed for reproducibility. Random by default.', default=None)
    prompt: str = Field(title='Prompt', default='')
    image_1: ImageRef = Field(default=ImageRef(), description='First input image')
    image_2: ImageRef = Field(default=ImageRef(), description='Second input image')
    image_3: ImageRef = Field(default=ImageRef(), description='Third input image (optional)')
    image_4: ImageRef = Field(default=ImageRef(), description='Fourth input image (optional)')
    image_5: ImageRef = Field(default=ImageRef(), description='Fifth input image (optional)')
    image_6: ImageRef = Field(default=ImageRef(), description='Sixth input image (optional)')
    image_7: ImageRef = Field(default=ImageRef(), description='Seventh input image (optional)')
    image_8: ImageRef = Field(default=ImageRef(), description='Eighth input image (optional)')
    image_9: ImageRef = Field(default=ImageRef(), description='Ninth input image (optional)')
    image_10: ImageRef = Field(default=ImageRef(), description='Tenth input image (optional)')
    max_width: int = Field(title='Max Width', description='Maximum width of the video', ge=256.0, le=768.0, default=512)
    max_height: int = Field(title='Max Height', description='Maximum height of the video', ge=256.0, le=768.0, default=512)
    interpolate: bool = Field(title='Interpolate', description='Enable 2x interpolation using FILM', default=False)
    negative_prompt: str = Field(title='Negative Prompt', description='Things you do not want to see in your video', default='')
    color_correction: bool = Field(title='Color Correction', description='If the colors are coming out strange, or if the colors between your input images are very different, disable this', default=True)

class Model(str, Enum):
    XL = 'xl'
    _576W = '576w'
    POTAT1 = 'potat1'
    ANIMOV_512X = 'animov-512x'


class Zeroscope_V2_XL(ReplicateNode):
    """Zeroscope V2 XL & 576w"""

    @classmethod
    def replicate_model_id(cls): return 'anotherjesse/zeroscope-v2-xl:9f747673945c62801b13b84701c783929c0ee784e4748ec062204894dda1a351'
    @classmethod
    def get_hardware(cls): return 'Nvidia A100 (40GB) GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/021a0a2e-b57b-4b45-a5c7-e29773aa5345/1mrNnh8.png', 'created_at': '2023-06-24T18:30:41.874899Z', 'description': 'Zeroscope V2 XL & 576w', 'github_url': 'https://github.com/anotherjesse/cog-text2video', 'license_url': 'https://github.com/anotherjesse/cog-text2video/blob/main/LICENSE', 'name': 'zeroscope-v2-xl', 'owner': 'anotherjesse', 'paper_url': 'https://huggingface.co/cerspense/zeroscope_v2_576w', 'run_count': 262043, 'url': 'https://replicate.com/anotherjesse/zeroscope-v2-xl', 'visibility': 'public', 'hardware': 'Nvidia A100 (40GB) GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    fps: int = Field(title='Fps', description='fps for the output video', default=8)
    seed: int | None = Field(title='Seed', description='Random seed. Leave blank to randomize the seed', default=None)
    model: Model = Field(description='Model to use', default='xl')
    width: int = Field(title='Width', description='Width of the output video', ge=256.0, default=576)
    height: int = Field(title='Height', description='Height of the output video', ge=256.0, default=320)
    prompt: str = Field(title='Prompt', description='Input prompt', default='An astronaut riding a horse')
    batch_size: int = Field(title='Batch Size', description='Batch size', ge=1.0, default=1)
    init_video: str | None = Field(title='Init Video', description='URL of the initial video (optional)', default=None)
    num_frames: int = Field(title='Num Frames', description='Number of frames for the output video', default=24)
    init_weight: float = Field(title='Init Weight', description='Strength of init_video', default=0.5)
    guidance_scale: float = Field(title='Guidance Scale', description='Guidance scale', ge=1.0, le=100.0, default=7.5)
    negative_prompt: str | None = Field(title='Negative Prompt', description='Negative prompt', default=None)
    remove_watermark: bool = Field(title='Remove Watermark', description='Remove watermark', default=False)
    num_inference_steps: int = Field(title='Num Inference Steps', description='Number of denoising steps', ge=1.0, le=500.0, default=50)

class Output_type(str, Enum):
    GREEN_SCREEN = 'green-screen'
    ALPHA_MASK = 'alpha-mask'
    FOREGROUND_MASK = 'foreground-mask'


class RobustVideoMatting(ReplicateNode):
    """extract foreground of a video"""

    @classmethod
    def replicate_model_id(cls): return 'arielreplicate/robust_video_matting:73d2128a371922d5d1abf0712a1d974be0e4e2358cc1218e4e34714767232bac'
    @classmethod
    def get_hardware(cls): return 'Nvidia T4 GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/1f92fd8f-2b90-4998-b5ae-1e23678ab004/showreel.gif', 'created_at': '2022-11-25T14:06:18.152759Z', 'description': 'extract foreground of a video', 'github_url': 'https://github.com/PeterL1n/RobustVideoMatting', 'license_url': 'https://github.com/PeterL1n/RobustVideoMatting/blob/master/LICENSE', 'name': 'robust_video_matting', 'owner': 'arielreplicate', 'paper_url': 'https://arxiv.org/abs/2108.11515', 'run_count': 44458, 'url': 'https://replicate.com/arielreplicate/robust_video_matting', 'visibility': 'public', 'hardware': 'Nvidia T4 GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    input_video: VideoRef = Field(default=VideoRef(), description='Video to segment.')
    output_type: Output_type = Field(default='green-screen')



class StableDiffusionInfiniteZoom(ReplicateNode):
    """Use Runway's Stable-diffusion inpainting model to create an infinite loop video"""

    @classmethod
    def replicate_model_id(cls): return 'arielreplicate/stable_diffusion_infinite_zoom:a2527c5074fc0cf9fa6015a40d75d080d1ddf7082fabe142f1ccd882c18fce61'
    @classmethod
    def get_hardware(cls): return 'Nvidia A100 (40GB) GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/0d244e1c-0f30-49ee-acb4-6a451c27f82f/infinite_zoom.gif', 'created_at': '2022-10-30T16:19:54.154872Z', 'description': "Use Runway's Stable-diffusion inpainting model to create an infinite loop video", 'github_url': 'https://github.com/ArielReplicate/stable-diffusion-infinite-zoom', 'license_url': 'https://github.com/ArielReplicate/stable-diffusion-infinite-zoom/blob/add-cog/LICENSE', 'name': 'stable_diffusion_infinite_zoom', 'owner': 'arielreplicate', 'paper_url': None, 'run_count': 35788, 'url': 'https://replicate.com/arielreplicate/stable_diffusion_infinite_zoom', 'visibility': 'public', 'hardware': 'Nvidia A100 (40GB) GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    prompt: str | None = Field(title='Prompt', description='Prompt', default=None)
    inpaint_iter: int = Field(title='Inpaint Iter', description="Number of iterations of pasting the image in it's center and inpainting the boarders", default=2)
    output_format: Output_format = Field(description='infinite loop gif or mp4 video', default='mp4')



class AnimateDiffIllusions(ReplicateNode):
    """Monster Labs' Controlnet QR Code Monster v2 For SD-1.5 on top of AnimateDiff Prompt Travel (Motion Module SD 1.5 v2)"""

    @classmethod
    def replicate_model_id(cls): return 'zsxkib/animatediff-illusions:b3ccb0101402aafd04bfea042950be606223e2abedbad93cf848bfffa072bb61'
    @classmethod
    def get_hardware(cls): return 'Nvidia A40 (Large) GPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/f6b87381-05b4-44d2-a1fd-f6bf45cf2580/output.gif', 'created_at': '2023-10-26T12:08:10.288839Z', 'description': "Monster Labs' Controlnet QR Code Monster v2 For SD-1.5 on top of AnimateDiff Prompt Travel (Motion Module SD 1.5 v2)", 'github_url': 'https://github.com/s9roll7/animatediff-cli-prompt-travel', 'license_url': 'https://github.com/s9roll7/animatediff-cli-prompt-travel/blob/main/LICENSE.md', 'name': 'animatediff-illusions', 'owner': 'zsxkib', 'paper_url': 'https://arxiv.org/abs/2307.04725', 'run_count': 8431, 'url': 'https://replicate.com/zsxkib/animatediff-illusions', 'visibility': 'public', 'hardware': 'Nvidia A40 (Large) GPU'}
    @classmethod
    def return_type(cls): return VideoRef


    loop: bool = Field(title='Loop', description="Flag to loop the video. Use when you have an 'infinitely' repeating video/gif ControlNet video", default=True)
    seed: int | None = Field(title='Seed', description='Seed for different images and reproducibility. Leave blank to randomise seed', default=None)
    steps: int = Field(title='Steps', description='Number of inference steps', ge=1.0, le=100.0, default=25)
    width: int = Field(title='Width', description='Width of generated video in pixels, must be divisable by 8', ge=64.0, le=2160.0, default=256)
    frames: int = Field(title='Frames', description='Length of the video in frames (playback is at 8 fps e.g. 16 frames @ 8 fps is 2 seconds)', ge=1.0, le=1024.0, default=128)
    height: int = Field(title='Height', description='Height of generated video in pixels, must be divisable by 8', ge=64.0, le=2160.0, default=384)
    context: int = Field(title='Context', description='Number of frames to condition on (default: max of <length> or 32). max for motion module v1 is 24', ge=1.0, le=32.0, default=16)
    clip_skip: int = Field(title='Clip Skip', description='Skip the last N-1 layers of the CLIP text encoder (lower values follow prompt more closely)', ge=1.0, le=6.0, default=2)
    scheduler: Scheduler = Field(description='Diffusion scheduler', default='k_dpmpp_sde')
    base_model: Base_model = Field(description="Choose the base model for animation generation. If 'CUSTOM' is selected, provide a custom model URL in the next parameter", default='majicmixRealistic_v5Preview')
    prompt_map: str = Field(title='Prompt Map', description="Prompt for changes in animation. Provide 'frame number : prompt at this frame', separate different prompts with '|'. Make sure the frame number does not exceed the length of video (frames)", default='')
    head_prompt: str = Field(title='Head Prompt', description='Primary animation prompt. If a prompt map is provided, this will be prefixed at the start of every individual prompt in the map', default='masterpiece, best quality, a haunting and detailed depiction of a ship at sea, battered by waves, ominous,((dark clouds:1.3)),distant lightning, rough seas, rain, silhouette of the ship against the stormy sky')
    tail_prompt: str = Field(title='Tail Prompt', description='Additional prompt that will be appended at the end of the main prompt or individual prompts in the map', default='')
    output_format: Output_format = Field(description="Output format of the video. Can be 'mp4' or 'gif'", default='mp4')
    guidance_scale: float = Field(title='Guidance Scale', description='Guidance Scale. How closely do we want to adhere to the prompt and its contents', ge=0.0, le=20.0, default=7.5)
    negative_prompt: str = Field(title='Negative Prompt', default='')
    controlnet_video: VideoRef = Field(default=VideoRef(), description='A short video/gif that will be used as the keyframes for QR Code Monster to use, Please note, all of the frames will be used as keyframes')
    film_interpolation: bool = Field(title='Film Interpolation', description='Whether to use FILM for between-frame interpolation (film-net.github.io)', default=True)
    prompt_fixed_ratio: float = Field(title='Prompt Fixed Ratio', description='Defines the ratio of adherence to the fixed part of the prompt versus the dynamic part (from prompt map). Value should be between 0 (only dynamic) to 1 (only fixed).', ge=0.0, le=1.0, default=0.5)
    custom_base_model_url: str = Field(title='Custom Base Model Url', description="Only used when base model is set to 'CUSTOM'. URL of the custom model to download if 'CUSTOM' is selected in the base model. Only downloads from 'https://civitai.com/api/download/models/' are allowed", default='')
    num_interpolation_steps: int = Field(title='Num Interpolation Steps', description='Number of steps to interpolate between animation frames', ge=1.0, le=50.0, default=3)
    enable_qr_code_monster_v2: bool = Field(title='Enable Qr Code Monster V2', description='Flag to enable QR Code Monster V2 ControlNet', default=True)
    playback_frames_per_second: int = Field(title='Playback Frames Per Second', ge=1.0, le=60.0, default=8)
    controlnet_conditioning_scale: float = Field(title='Controlnet Conditioning Scale', description='Strength of ControlNet. The outputs of the ControlNet are multiplied by `controlnet_conditioning_scale` before they are added to the residual in the original UNet', default=0.18)
    qr_code_monster_v2_guess_mode: bool = Field(title='Qr Code Monster V2 Guess Mode', description='Flag to enable guess mode (un-guided) for QR Code Monster V2 ControlNet', default=False)
    qr_code_monster_v2_preprocessor: bool = Field(title='Qr Code Monster V2 Preprocessor', description='Flag to pre-process keyframes for QR Code Monster V2 ControlNet', default=True)



class AudioToWaveform(ReplicateNode):
    """Create a waveform video from audio"""

    @classmethod
    def replicate_model_id(cls): return 'fofr/audio-to-waveform:116cf9b97d0a117cfe64310637bf99ae8542cc35d813744c6ab178a3e134ff5a'
    @classmethod
    def get_hardware(cls): return 'CPU'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/5d5cad9c-d4ba-44e1-8c4f-dc08648bbf5e/fofr_a_waveform_bar_chart_video_e.png', 'created_at': '2023-06-13T15:26:38.672021Z', 'description': 'Create a waveform video from audio', 'github_url': 'https://github.com/fofr/audio-to-waveform', 'license_url': 'https://github.com/fofr/audio-to-waveform/blob/main/LICENSE', 'name': 'audio-to-waveform', 'owner': 'fofr', 'paper_url': 'https://gradio.app/docs/#make_waveform', 'run_count': 368988, 'url': 'https://replicate.com/fofr/audio-to-waveform', 'visibility': 'public', 'hardware': 'CPU'}
    @classmethod
    def return_type(cls): return VideoRef


    audio: AudioRef = Field(default=AudioRef(), description='Audio file to create waveform from')
    bg_color: str = Field(title='Bg Color', description='Background color of waveform', default='#000000')
    fg_alpha: float = Field(title='Fg Alpha', description='Opacity of foreground waveform', default=0.75)
    bar_count: int = Field(title='Bar Count', description='Number of bars in waveform', default=100)
    bar_width: float = Field(title='Bar Width', description='Width of bars in waveform. 1 represents full width, 0.5 represents half width, etc.', default=0.4)
    bars_color: str = Field(title='Bars Color', description='Color of waveform bars', default='#ffffff')
    caption_text: str = Field(title='Caption Text', description='Caption text for the video', default='')


import io

import numpy as np
import pytest
from PIL import Image

from controlnet_aux.processor import MODELS, Processor


@pytest.fixture(params=[
    'scribble_hed',
    'softedge_hed',
    'scribble_hedsafe',
    'softedge_hedsafe',
    'depth_midas',
    'mlsd',
    'openpose',
    'openpose_hand',
    'openpose_face',
    'openpose_faceonly',
    'openpose_full',
    'scribble_pidinet',
    'softedge_pidinet',
    'scribble_pidsafe',
    'softedge_pidsafe',
    'normal_bae',
    'lineart_coarse',
    'lineart_realistic',
    'lineart_anime',
    'canny',
    'shuffle',
    'depth_zoe',
    'depth_leres',
    'depth_leres++',
    'mediapipe_face',
    'tile'
])
def processor(request):
    return Processor(request.param)


def test_processor_init(processor):
    assert isinstance(processor.processor, MODELS[processor.processor_id]['class'])
    assert isinstance(processor.params, dict)


def test_processor_call(processor):
    # Load test image
    with open('test_image.png', 'rb') as f:
        image_bytes = f.read()
    image = Image.open(io.BytesIO(image_bytes))

    # Output size
    resolution = 512
    W, H = image.size
    H = float(H)
    W = float(W)
    k = float(resolution) / min(H, W)
    H *= k
    W *= k
    H = int(np.round(H / 64.0)) * 64
    W = int(np.round(W / 64.0)) * 64

    # Test processing
    processed_image = processor(image)
    assert isinstance(processed_image, Image.Image)
    assert processed_image.size == (W, H)


def test_processor_call_bytes(processor):
    # Load test image
    with open('test_image.png', 'rb') as f:
        image_bytes = f.read()

    # Test processing
    processed_image_bytes = processor(image_bytes, to_pil=False)
    assert isinstance(processed_image_bytes, bytes)
    assert len(processed_image_bytes) > 0


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Blend(GraphNode):
    image1: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The first image to blend.')
    image2: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The second image to blend.')
    alpha: float | GraphNode | tuple[GraphNode, str] = Field(default=0.5, description='The mix ratio.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.Blend"



class Composite(GraphNode):
    image1: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The first image to composite.')
    image2: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The second image to composite.')
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The mask to composite with.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.Composite"



class ImageToTensor(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The input image to convert to a tensor. The image should have either 1 (grayscale), 3 (RGB), or 4 (RGBA) channels.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.ImageToTensor"



class Paste(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to paste into.')
    paste: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to paste.')
    left: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The left coordinate.')
    top: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The top coordinate.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.Paste"



class SaveImage(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description=None)
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to save the image in.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='%Y-%m-%d_%H-%M-%S.png', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.image.SaveImage"



class TensorToImage(GraphNode):
    tensor: Tensor | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The input tensor to convert to an image. Should have either 1, 3, or 4 channels.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.TensorToImage"





import os
os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

import torch

from custom_detectron2.config import get_cfg
from custom_detectron2.projects.deeplab import add_deeplab_config
from custom_detectron2.data import MetadataCatalog

from custom_oneformer import (
    add_oneformer_config,
    add_common_config,
    add_swin_config,
    add_dinat_config,
)

from custom_oneformer.demo.defaults import DefaultPredictor
from custom_oneformer.demo.visualizer import Visualizer, ColorMode


def make_detectron2_model(config_path, ckpt_path):
    cfg = get_cfg()
    add_deeplab_config(cfg)
    add_common_config(cfg)
    add_swin_config(cfg)
    add_oneformer_config(cfg)
    add_dinat_config(cfg)
    cfg.merge_from_file(config_path)
    cfg.MODEL.WEIGHTS = ckpt_path
    cfg.freeze()
    metadata = MetadataCatalog.get(cfg.DATASETS.TEST_PANOPTIC[0] if len(cfg.DATASETS.TEST_PANOPTIC) else "__unused")
    return DefaultPredictor(cfg), metadata


def semantic_run(img, predictor, metadata):
    predictions = predictor(img[:, :, ::-1], "semantic")  # Predictor of OneFormer must use BGR image !!!
    visualizer_map = Visualizer(img, is_img=False, metadata=metadata, instance_mode=ColorMode.IMAGE)
    out_map = visualizer_map.draw_sem_seg(predictions["sem_seg"].argmax(dim=0).cpu(), alpha=1, is_text=False).get_image()
    return out_map


from nodetool.metadata.types import OutputSlot
from nodetool.workflows.base_node import BaseNode, type_metadata
from nodetool.workflows.property import Property
from nodetool.workflows.read_graph import read_graph
from nodetool.api.types.graph import Graph as APIGraph
import json
from nodetool.workflows.run_workflow import run_workflow
from nodetool.workflows.run_job_request import RunJobRequest
from nodetool.common.environment import Environment
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.workflows.base_node import InputNode, OutputNode
from nodetool.workflows.graph import Graph
from typing import Any, Type

from nodetool.workflows.types import NodeProgress, NodeUpdate, WorkflowUpdate


def properties_from_input_nodes(nodes: list[BaseNode]):
    return [
        Property(
            name=node.name,
            title=node.name,
            default=node.value,  # type: ignore
            min=node.min if hasattr(node, "min") else None,  # type: ignore
            max=node.max if hasattr(node, "max") else None,  # type: ignore
            type=node.properties_dict()["value"].type,
        )
        for node in nodes
        if isinstance(node, InputNode)
    ]


class WorkflowNode(BaseNode):
    inputs: dict[str, Any] = {}

    def __init__(self, **kwargs):
        id = kwargs.pop("_id", "")
        ui_properties = kwargs.pop("_ui_properties", {})
        super().__init__(id=id, ui_properties=ui_properties)
        self.inputs = kwargs

    def assign_property(self, name: str, value: Any):
        self.inputs[name] = value

    @classmethod
    def get_workflow_file(cls) -> str:
        return ""

    @classmethod
    def read_workflow(cls) -> dict:
        workflow_file = cls.get_workflow_file()
        if workflow_file == "":
            return {}
        if not hasattr(cls, "workflow_json"):
            with open(workflow_file, "r") as f:
                cls.workflow_json = json.load(f)
        return cls.workflow_json

    @classmethod
    def properties(cls):
        graph = cls.load_graph()
        return properties_from_input_nodes(graph.nodes)

    @classmethod
    def load_workflow(cls):
        return read_graph(cls.read_workflow())

    @classmethod
    def load_graph(cls):
        edges, nodes = cls.load_workflow()
        return Graph.from_dict(
            {
                "nodes": [node.model_dump() for node in nodes],
                "edges": [edge.model_dump() for edge in edges],
            }
        )

    @classmethod
    def outputs(cls):
        graph = cls.load_graph()
        return [
            OutputSlot(type=type_metadata(output.return_type()), name=output.name)  # type: ignore
            for output in graph.outputs()
        ]

    def get_api_graph(self) -> APIGraph:
        edges, nodes = self.load_workflow()
        return APIGraph(edges=edges, nodes=nodes)

    async def process(self, context: ProcessingContext):
        logs = ""

        req = RunJobRequest(
            user_id=context.user_id,
            auth_token=context.auth_token,
            graph=self.get_api_graph(),
            params=self.inputs or {},
        )
        output = {}
        for msg_json in run_workflow(req):
            msg = json.loads(msg_json)
            if msg["type"] == "node_progress":
                context.post_message(
                    NodeProgress(
                        node_id=self._id,
                        progress=msg["progress"],
                        total=msg["total"],
                    )
                )
            if msg["type"] == "node_update":
                logs += f"{msg['node_name']} -> {msg['status']}\n"
                context.post_message(
                    NodeUpdate(
                        node_id=self._id,
                        node_name=self.get_title(),
                        status="running",
                        logs=logs,
                    )
                )
            if msg["type"] == "error":
                raise Exception(msg["error"])
            if msg["type"] == "workflow_update":
                output = msg["result"]

        return output
